#
# File: TorsionLibraryAlerts.py
# Author: Manish Sud <msud@san.rr.com>
#
# Collaborator: Pat Walters
#
# Acknowledgments: Wolfgang Guba, Patrick Penner, and Levi Pierce
#
# Copyright (C) 2024 Manish Sud. All rights reserved.
#
# This module uses the Torsion Library jointly developed by the University
# of Hamburg, Center for Bioinformatics, Hamburg, Germany and
# F. Hoffmann-La-Roche Ltd., Basel, Switzerland.
#
# This file is part of MayaChemTools.
#
# MayaChemTools is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.
#
# MayaChemTools is distributed in the hope that it will be useful, but without
# any warranty; without even the implied warranty of merchantability of fitness
# for a particular purpose.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with MayaChemTools; if not, see <http://www.gnu.org/licenses/> or
# write to the Free Software Foundation Inc., 59 Temple Place, Suite 330,
# Boston, MA, 02111-1307, USA.
#

import os
import re
import math
import numpy as np

from rdkit import Chem
from rdkit.Chem import rdMolTransforms

from . import TorsionAlertsUtil

class TorsionLibraryAlerts():
    def __init__(self, AlertsMode = "Red", MinAlertsCount = 1, NitrogenLonePairAllowHydrogenNbrs = True, NitrogenLonePairPlanarityTolerance = 1.0, RotBondsSMARTSMode = "SemiStrict", RotBondsSMARTSPattern = None, TorsionLibraryFilePath = "auto"):
        """Identify strained molecules from an input file for torsion library [ Ref 146, 152, 159 ]
        alerts by matching rotatable bonds against SMARTS patterns specified for torsion
        rules in a torsion library file, The molecules must have 3D coordinates. The default
        torsion library file, TorsionLibrary.xml, is available in the directory containing this file.
        
        The data in torsion library file is organized in a hierarchical manner. It consists
        of one generic class and six specific classes at the highest level. Each class
        contains multiple subclasses corresponding to named functional groups or
        substructure patterns. The subclasses consist of torsion rules sorted from
        specific to generic torsion patterns. The torsion rule, in turn, contains a list
        of peak values for torsion angles and two tolerance values. A pair of tolerance
        values define torsion bins around a torsion peak value. For example:
             
            <library>
                <hierarchyClass name="GG" id1="G" id2="G">
                ...
                </hierarchyClass>
                <hierarchyClass name="CO" id1="C" id2="O">
                    <hierarchySubClass name="Ester bond I" smarts="O=[C:2][O:3]">
                        <torsionRule smarts="[O:1]=[C:2]!@[O:3]~[CH0:4]">
                            <angleList>
                                <angle value="0.0" tolerance1="20.00"
                                 tolerance2="25.00" score="56.52"/>
                            </angleList>
                        </torsionRule>
                        ...
                    ...
                 ...
                </hierarchyClass>
                <hierarchyClass name="NC" id1="N" id2="C">
                 ...
                </hierarchyClass>
                <hierarchyClass name="SN" id1="S" id2="N">
                ...
                </hierarchyClass>
                <hierarchyClass name="CS" id1="C" id2="S">
                ...
                </hierarchyClass>
                <hierarchyClass name="CC" id1="C" id2="C">
                ...
                </hierarchyClass>
                <hierarchyClass name="SS" id1="S" id2="S">
                 ...
                </hierarchyClass>
            </library>
            
        The rotatable bonds in a 3D molecule are identified using a default SMARTS pattern.
        A custom SMARTS pattern may be optionally specified to detect rotatable bonds.
        Each rotatable bond is matched to a torsion rule in the torsion library and
        assigned one of the following three alert categories: Green, Orange or Red. The 
        rotatable bond is marked Green or Orange for the measured angle of the torsion
        pattern within the first or second tolerance bins around a torsion peak.
        Otherwise, it's marked Red implying that the measured angle is not observed in
        the structure databases employed to generate the torsion library.

        Arguments:
            AlertsMode(str): Torsion library alert types to use for issuing
                alerts about molecules.
            MinAlertsCount(int): Minimum number of alerts allowed in a molecule.
            NitrogenLonePairAllowHydrogenNbrs (bool): Use hydrogen neighbors
                attached to nitrogen during the determination of its planarity.
            NitrogenLonePairPlanarityTolerance (float): Angle tolerance in
                degrees allowed for nitrogen to be considered coplanar with its
                three neighbors.
            RotBondsSMARTSMode (str): SMARTS pattern to use for identifying
                rotatable bonds in a molecule. Possible values: NonStrict,
                SemiStrict, Strict or Specify.
            RotBondsSMARTSPattern (str):SMARTS pattern for identifying rotatable
                bonds. This paramater is only valid for 'Specify' value
                'RotBondsSMARTSMode'.
            TorsionLibraryFilePath (str): A XML file name containing data for
                torsion library.

        Returns:
            object: An instantiated class object.

        Notes:
            The following sections provide additional details for the parameters.
            
            AlertsMode: Torsion library alert types to use for issuing alerts about
            molecules containing rotatable bonds marked with Green, Orange, or
            Red alerts. Possible values: Red or RedAndOrange.
            
            MinAlertsCount: Minimum number of rotatable bond alerts allowed in a
            molecule.
            
            NitrogenLonePairAllowHydrogenNbrs, NitrogenLonePairPlanarityTolerance:
            These parameters are used during the matching of torsion rules containing
            'N_lp' in their SMARTS patterns. The 'allowHydrogensNbrs' allows the use
            hydrogen neighbors attached to nitrogen during the determination of its
            planarity. The 'planarityTolerance' in degrees represents the tolerance
            allowed for nitrogen to be considered coplanar with its three neighbors.
            
            The torsion rules containing 'N_lp' in their SMARTS patterns are categorized
            into the following two types of rules:
                
                TypeOne:  
                
                [CX4:1][CX4H2:2]!@[NX3;"N_lp":3][CX4:4]
                [C:1][CX4H2:2]!@[NX3;"N_lp":3][C:4]
                ... ... ...
                
                TypeTwo:  
                
                [!#1:1][CX4:2]!@[NX3;"N_lp":3]
                [C:1][$(S(=O)=O):2]!@["N_lp":3]
                ... ... ...
                
            The torsions are matched to torsion rules containing 'N_lp' using specified
            SMARTS patterns without the 'N_lp' along with additional constraints using
            the following methodology:
                
                TypeOne:  
                
                . SMARTS pattern must contain four mapped atoms and the third
                    mapped atom must be a nitrogen matched with 'NX3:3'
                . Nitrogen atom must have 3 neighbors. The 'allowHydrogens'
                    parameter controls inclusion of hydrogens as its neighbors.
                . Nitrogen atom and its 3 neighbors must be coplanar.
                    'planarityTolerance' parameter provides tolerance in degrees
                    for nitrogen to be considered coplanar with its 3 neighbors.
                
                TypeTwo:  
                
                . SMARTS pattern must contain three mapped atoms and the third
                    mapped atom must be a nitrogen matched with 'NX3:3'. The 
                    third mapped atom may contain only 'N_lp:3' The missing 'NX3'
                    is automatically detected.
                . Nitrogen atom must have 3 neighbors. 'allowHydrogens'
                    parameter controls inclusion of hydrogens as neighbors.
                . Nitrogen atom and its 3 neighbors must not be coplanar.
                    'planarityTolerance' parameter provides tolerance in degrees
                    for nitrogen to be considered coplanar with its 3 neighbors.
                . Nitrogen lone pair position equivalent to VSEPR theory is
                    determined based on the position of nitrogen and its neighbors.
                    A vector normal to 3 nitrogen neighbors is calculated and added
                    to the coordinates of nitrogen atom to determine the approximate
                    position of the lone pair. It is used as the fourth position to
                    calculate the torsion angle.
                 
            RotBondsSMARTSMode: SMARTS pattern to use for identifying rotatable bonds in
            a molecule for matching against torsion rules in the torsion library. Possible
            values: NonStrict, SemiStrict, Strict or Specify. The rotatable bond SMARTS
            matches are filtered to ensure that each atom in the rotatable bond is attached
            to at least two heavy atoms.
            
            The following SMARTS patterns are used to identify rotatable bonds for
            different modes:
                
                NonStrict: [!$(*#*)&!D1]-&!@[!$(*#*)&!D1]
                
                SemiStrict:
                [!$(*#*)&!D1&!$(C(F)(F)F)&!$(C(Cl)(Cl)Cl)&!$(C(Br)(Br)Br)
                &!$(C([CH3])([CH3])[CH3])]-!@[!$(*#*)&!D1&!$(C(F)(F)F)
                &!$(C(Cl)(Cl)Cl)&!$(C(Br)(Br)Br)&!$(C([CH3])([CH3])[CH3])]
                
                Strict:
                [!$(*#*)&!D1&!$(C(F)(F)F)&!$(C(Cl)(Cl)Cl)&!$(C(Br)(Br)Br)
                &!$(C([CH3])([CH3])[CH3])&!$([CD3](=[N,O,S])-!@[#7,O,S!D1])
                &!$([#7,O,S!D1]-!@[CD3]=[N,O,S])&!$([CD3](=[N+])-!@[#7!D1])
                &!$([#7!D1]-!@[CD3]=[N+])]-!@[!$(*#*)&!D1&!$(C(F)(F)F)
                &!$(C(Cl)(Cl)Cl)&!$(C(Br)(Br)Br)&!$(C([CH3])([CH3])[CH3])]
                
            The 'NonStrict' and 'Strict' SMARTS patterns are available in RDKit. The 
            'NonStrict' SMARTS pattern corresponds to original Daylight SMARTS
            specification for rotatable bonds. The 'SemiStrict' SMARTS pattern is 
            derived from 'Strict' SMARTS pattern.
            
            TorsionLibraryFilePath: Specify a XML file name containing data for
            torsion library hierarchy or use default file, TorsionLibrary.xml, available in
            the directory containing this file.

        """

        self._ProcessTorsionLibraryAlertsParameters(AlertsMode, MinAlertsCount, NitrogenLonePairAllowHydrogenNbrs, NitrogenLonePairPlanarityTolerance, RotBondsSMARTSMode, RotBondsSMARTSPattern, TorsionLibraryFilePath)

        self._InitializeTorsionLibraryAlerts()


    def _ProcessTorsionLibraryAlertsParameters(self, AlertsMode, MinAlertsCount, NitrogenLonePairAllowHydrogenNbrs, NitrogenLonePairPlanarityTolerance, RotBondsSMARTSMode, RotBondsSMARTSPattern, TorsionLibraryFilePath):
        """Process torsion library alerts paramaters. """

        # Process AltertsMode paramater...
        self._ProcessAlertsModeParameter(AlertsMode)

        # Process MinAlertsCount...
        if MinAlertsCount < 1:
            raise ValueError("The value, %s, specified for AltertsMinCount parameter is not valid. Supported value: >=1" % MinAlertsCount)
        self.MinAlertsCount = MinAlertsCount

        # Process nitrogen lone pair paramaters...
        self.NitrogenLonePairAllowHydrogenNbrs = NitrogenLonePairAllowHydrogenNbrs
        if NitrogenLonePairPlanarityTolerance < 0:
            raise ValueError("The value, %s, specified for NitrogenLonePairPlanarityTolerance parameter is not valid. Supported value: >= 0" % NitrogenLonePairPlanarityTolerance)
        self.NitrogenLonePairPlanarityTolerance = NitrogenLonePairPlanarityTolerance

        # Process RotBondsSMARTSMode and RotBondsSMARTSPattern parameters...
        self._ProcessRotBondsParameters(RotBondsSMARTSMode, RotBondsSMARTSPattern)

        # Process TorsionLibraryFilePath parameter...
        self._ProcessTorsionLibraryFilePathParameter(TorsionLibraryFilePath)


    def _InitializeTorsionLibraryAlerts(self):
        """Initialize tosion library alerts."""

        # Retrieve and setup torsion library info for matching rotatable bonds...
        TorsionLibraryInfo = {}

        TorsionLibElementTree = TorsionAlertsUtil.RetrieveTorsionLibraryInfo(self.TorsionLibraryFilePath)
        TorsionLibraryInfo["TorsionLibElementTree"] = TorsionLibElementTree

        TorsionAlertsUtil.SetupTorsionLibraryInfoForMatchingRotatableBonds(TorsionLibraryInfo)

        self.TorsionLibraryInfo = TorsionLibraryInfo


    def _ProcessAlertsModeParameter(self, AlertsMode):
        """Process AlertsMode parameter. """

        SpecifiedAlertsModeList = []
        if re.match("^Red$", AlertsMode, re.I):
            SpecifiedAlertsModeList.append("Red")
        elif re.match("^RedAndOrange$", AlertsMode, re.I):
            SpecifiedAlertsModeList.append("Red")
            SpecifiedAlertsModeList.append("Orange")
        else:
            raise ValueError("Invalid value, %s, specified for AlertsMode parameter. Valid values: Red, or RedAndOrange" % (AlertsMode))

        self.AltersMode = AlertsMode
        self.SpecifiedAlertsModeList = SpecifiedAlertsModeList


    def _ProcessRotBondsParameters(self, RotBondsSMARTSMode, RotBondsSMARTSPattern):
        """Process  RotBondsSMARTSMode and RotBondsSMARTSPattern parameters. """

        if re.match("^NonStrict$", RotBondsSMARTSMode, re.I):
            RotBondsSMARTSPattern = "[!$(*#*)&!D1]-&!@[!$(*#*)&!D1]"
        elif re.match("^SemiStrict$", RotBondsSMARTSMode, re.I):
            RotBondsSMARTSPattern = "[!$(*#*)&!D1&!$(C(F)(F)F)&!$(C(Cl)(Cl)Cl)&!$(C(Br)(Br)Br)&!$(C([CH3])([CH3])[CH3])]-!@[!$(*#*)&!D1&!$(C(F)(F)F)&!$(C(Cl)(Cl)Cl)&!$(C(Br)(Br)Br)&!$(C([CH3])([CH3])[CH3])]"
        elif re.match("^Strict$", RotBondsSMARTSMode, re.I):
            RotBondsSMARTSPattern = "[!$(*#*)&!D1&!$(C(F)(F)F)&!$(C(Cl)(Cl)Cl)&!$(C(Br)(Br)Br)&!$(C([CH3])([CH3])[CH3])&!$([CD3](=[N,O,S])-!@[#7,O,S!D1])&!$([#7,O,S!D1]-!@[CD3]=[N,O,S])&!$([CD3](=[N+])-!@[#7!D1])&!$([#7!D1]-!@[CD3]=[N+])]-!@[!$(*#*)&!D1&!$(C(F)(F)F)&!$(C(Cl)(Cl)Cl)&!$(C(Br)(Br)Br)&!$(C([CH3])([CH3])[CH3])]"
        elif re.match("^Specify$", RotBondsSMARTSMode, re.I):
            if RotBondsSMARTSPattern is None:
                raise ValueError("The RotBondsSMARTSPattern parameter value must be specified during Specify value for RotBondsSMARTSMode parameter.")
            RotBondsSMARTSPattern = RotBondsSMARTSPattern.strip()
            if not len(RotBondsSMARTSPattern):
                raise ValueError("Empty value specified for  RotBondsSMARTSPattern parameter.")
        else:
            raise ValueError("Invalid value, %s, specified for RotBondsSMARTSMode parameter. Valid values: NonStrict, SemiStrict, Strict or Specify" % (RotBondsSMARTSMode))

        RotBondsPatternMol = Chem.MolFromSmarts(RotBondsSMARTSPattern)
        if RotBondsPatternMol is None:
            if re.match("Specify", RotBondsSMARTSMode, re.I):
                raise ValueError("Failed to create rotatable bonds pattern molecule. The rotatable bonds SMARTS pattern, \"%s\", specified using RotBondsSMARTSPattern parameter is not valid." % (RotBondsSMARTSPattern))
            else:
                raise ValueError("Failed to create rotatable bonds pattern molecule. The default rotatable bonds SMARTS pattern, \"%s\", used for %s value of RotBondsSMARTSMode parameter is not valid." % (RotBondsSMARTSPattern, RotBondsSMARTSMode))

        self.RotBondsSMARTSMode = RotBondsSMARTSMode
        self.RotBondsSMARTSPattern = RotBondsSMARTSPattern
        self.RotBondsPatternMol = RotBondsPatternMol


    def _ProcessTorsionLibraryFilePathParameter(self, TorsionLibraryFilePath):
        """ Process torsion library path parameter. """

        # TorsionLibraryFilePath parameter...
        if re.match("^auto$", TorsionLibraryFilePath):
            TorsionLibraryFile = "TorsionLibrary.xml"
            TorsionLibraryFilePath = os.path.join(os.path.dirname(os.path.abspath(__file__)), TorsionLibraryFile)
            if not os.path.isfile(TorsionLibraryFilePath):
                raise ValueError("The default torsion alerts library file %s doesn't exist." % TorsionLibraryFilePath)
        else:
            if not os.path.isfile(TorsionLibraryFilePath):
                raise ValueError("The file specified, %s, for parameter TorsionLibraryFilePath doesn't exist." % TorsionLibraryFilePath)
            TorsionLibraryFilePath = os.path.abspath(TorsionLibraryFilePath)

        self.TorsionLibraryFilePath = TorsionLibraryFilePath


    def GetTorsionLibraryFilePath(self):
        """Get torsion strain library file path.

        Arguments:
            Nothing.

        Returns:
            FilePath (str): Torsion strain library path.

        """

        return self.TorsionLibraryFilePath


    def ListTorsionLibraryInfo(self):
        """List torsion strain library information.

        Arguments:
            Nothing.

        Returns:
            Nothing. The torsion library information is printed.

        """

        return TorsionAlertsUtil.ListTorsionLibraryInfo(self.TorsionLibraryInfo["TorsionLibElementTree"])


    def IdentifyTorsionLibraryAlertsForRotatableBonds(self, Mol):
        """Identify torsion library alerts for a molecule by matching rotatable bonds
        against SMARTS patterns specified for torsion rules in torsion energy library
        file.
 
        Arguments:
            Mol (object): RDKit molecule object.

        Returns:
            bool: True - Molecule contains strained torsions; False - Molecule
                contains no strained torsions or rotatable bonds. 
            dict or None: Torsion alerts information regarding matching of
                rotatable bonds to torsion strain library.

        Examples:

            from TorsionAlerts.TorsionLibraryAlerts import TorsionLibraryAlerts
            
            LibraryAlerts = TorsionLibraryAlerts()
            AlertsStatus, AlertsInfo = LibraryAlerts.
                IdentifyTorsionLibraryAlertsForRotatableBonds(RDKitMol)
            
            # List of rotatable bond IDs...
            RotatableBondIDs = AlertsInfo["IDs"]
            
            # Dictionaries containing information for rotatable bonds by using
            # bond ID as key...
            for ID in AlertsInfo["IDs"]:
                MatchStatus = AlertsInfo["MatchStatus"][ID]
                AlertTypes = AlertsInfo["AlertTypes"][ID]
                AtomIndices = AlertsInfo["AtomIndices"][ID]
                TorsionAtomIndices = AlertsInfo["TorsionAtomIndices"][ID]
                TorsionAngles = AlertsInfo["TorsionAngles"][ID]
                TorsionAngleViolations = AlertsInfo["TorsionAngleViolations"][ID]
                HierarchyClassNames = AlertsInfo["HierarchyClassNames"][ID]
                HierarchySubClassNames = AlertsInfo["HierarchySubClassNames"][ID]
                TorsionRuleNodeID = AlertsInfo["TorsionRuleNodeID"][ID]
                TorsionRulePeaks = AlertsInfo["TorsionRulePeaks"][ID]
                TorsionRuleTolerances1 = AlertsInfo["TorsionRuleTolerances1"][ID]
                TorsionRuleTolerances2 = AlertsInfo["TorsionRuleTolerances2"][ID]
                TorsionRuleSMARTS = AlertsInfo["TorsionRuleSMARTS"][ID]
                Count = AlertsInfo["Count"][ID]

        """

        # Identify rotatable bonds...
        RotBondsStatus, RotBondsInfo = TorsionAlertsUtil.IdentifyRotatableBondsForTorsionLibraryMatch(self.TorsionLibraryInfo, Mol, self.RotBondsPatternMol)
        if not RotBondsStatus:
            return (False, None)

        # Identify alerts for rotatable bonds...
        RotBondsAlertsStatus, RotBondsAlertsInfo = self._MatchRotatableBondsToTorsionLibrary(Mol, RotBondsInfo)

        return (RotBondsAlertsStatus, RotBondsAlertsInfo)


    def _MatchRotatableBondsToTorsionLibrary(self, Mol, RotBondsInfo):
        """Match rotatable bond to torsion library."""

        # Initialize...
        RotBondsAlertsInfo = self._InitializeRotatableBondsAlertsInfo()

        # Match rotatable bonds to torsion library...
        for ID in RotBondsInfo["IDs"]:
            AtomIndices = RotBondsInfo["AtomIndices"][ID]
            HierarchyClass = RotBondsInfo["HierarchyClass"][ID]

            MatchStatus, MatchInfo = self._MatchRotatableBondToTorsionLibrary(Mol, AtomIndices, HierarchyClass)

            if MatchInfo is None or len(MatchInfo) == 0:
                AlertType, TorsionAtomIndices, TorsionAngle, TorsionAngleViolation, HierarchyClassName, HierarchySubClassName, TorsionRuleNodeID, TorsionRulePeaks, TorsionRuleTolerances1, TorsionRuleTolerances2, TorsionRuleSMARTS = [None] * 11
            else:
                AlertType, TorsionAtomIndices, TorsionAngle, TorsionAngleViolation, HierarchyClassName, HierarchySubClassName, TorsionRuleNodeID, TorsionRulePeaks, TorsionRuleTolerances1, TorsionRuleTolerances2, TorsionRuleSMARTS = MatchInfo

            # Track alerts information...
            RotBondsAlertsInfo["IDs"].append(ID)
            RotBondsAlertsInfo["MatchStatus"][ID] = MatchStatus
            RotBondsAlertsInfo["AlertTypes"][ID] = AlertType
            RotBondsAlertsInfo["AtomIndices"][ID] = AtomIndices
            RotBondsAlertsInfo["TorsionAtomIndices"][ID] = TorsionAtomIndices
            RotBondsAlertsInfo["TorsionAngles"][ID] = TorsionAngle
            RotBondsAlertsInfo["TorsionAngleViolations"][ID] = TorsionAngleViolation
            RotBondsAlertsInfo["HierarchyClassNames"][ID] = HierarchyClassName
            RotBondsAlertsInfo["HierarchySubClassNames"][ID] = HierarchySubClassName
            RotBondsAlertsInfo["TorsionRuleNodeID"][ID] = TorsionRuleNodeID
            RotBondsAlertsInfo["TorsionRulePeaks"][ID] = TorsionRulePeaks
            RotBondsAlertsInfo["TorsionRuleTolerances1"][ID] = TorsionRuleTolerances1
            RotBondsAlertsInfo["TorsionRuleTolerances2"][ID] = TorsionRuleTolerances2
            RotBondsAlertsInfo["TorsionRuleSMARTS"][ID] = TorsionRuleSMARTS

            #  Count alert types...
            if AlertType is not None:
                if AlertType not in RotBondsAlertsInfo["Count"]:
                    RotBondsAlertsInfo["Count"][AlertType] = 0
                RotBondsAlertsInfo["Count"][AlertType] += 1

        # Setup alert status for rotatable bonds...
        RotBondsAlertsStatus = False
        AlertsCount = 0
        for ID in RotBondsInfo["IDs"]:
            if RotBondsAlertsInfo["AlertTypes"][ID] in self.SpecifiedAlertsModeList:
                AlertsCount += 1
                if AlertsCount >= self.MinAlertsCount:
                    RotBondsAlertsStatus = True
                    break

        return (RotBondsAlertsStatus, RotBondsAlertsInfo)


    def _InitializeRotatableBondsAlertsInfo(self, ):
        """Initialize alerts information for rotatable bonds."""

        RotBondsAlertsInfo = {}
        RotBondsAlertsInfo["IDs"] = []

        for DataLabel in ["MatchStatus", "AlertTypes", "AtomIndices", "TorsionAtomIndices", "TorsionAngles", "TorsionAngleViolations", "HierarchyClassNames", "HierarchySubClassNames", "TorsionRuleNodeID", "TorsionRulePeaks", "TorsionRuleTolerances1", "TorsionRuleTolerances2", "TorsionRuleSMARTS", "Count"]:
            RotBondsAlertsInfo[DataLabel] = {}

        return RotBondsAlertsInfo


    def _MatchRotatableBondToTorsionLibrary(self, Mol, RotBondAtomIndices, RotBondHierarchyClass):
        """Match rotatable bond to torsion library."""

        if TorsionAlertsUtil.IsSpecificHierarchyClass(self.TorsionLibraryInfo, RotBondHierarchyClass):
            MatchStatus, MatchInfo = self._MatchRotatableBondAgainstSpecificHierarchyClass(Mol, RotBondAtomIndices, RotBondHierarchyClass)
            if not MatchStatus:
                MatchStatus, MatchInfo = self._MatchRotatableBondAgainstGenericHierarchyClass(Mol, RotBondAtomIndices, RotBondHierarchyClass)
        else:
            MatchStatus, MatchInfo = self._MatchRotatableBondAgainstGenericHierarchyClass(Mol, RotBondAtomIndices, RotBondHierarchyClass)

        return (MatchStatus, MatchInfo)


    def _MatchRotatableBondAgainstSpecificHierarchyClass(self, Mol, RotBondAtomIndices, RotBondHierarchyClass):
        """Match rotatable bond against a specific hierarchy class."""

        TorsionLibraryInfo = self.TorsionLibraryInfo

        HierarchyClassElementNode = None
        if RotBondHierarchyClass in TorsionLibraryInfo["SpecificClasses"]["ElementNode"]:
            HierarchyClassElementNode = TorsionLibraryInfo["SpecificClasses"]["ElementNode"][RotBondHierarchyClass]

        if HierarchyClassElementNode is None:
            return (False, None, None, None)

        TorsionAlertsUtil.TrackHierarchyClassElementNode(TorsionLibraryInfo, HierarchyClassElementNode)
        MatchStatus, MatchInfo = self._ProcessElementForRotatableBondMatch(Mol, RotBondAtomIndices, HierarchyClassElementNode)
        TorsionAlertsUtil.RemoveLastHierarchyClassElementNodeFromTracking(TorsionLibraryInfo)

        return (MatchStatus, MatchInfo)


    def _MatchRotatableBondAgainstGenericHierarchyClass(self, Mol, RotBondAtomIndices, RotBondHierarchyClass):
        """Match rotatable bond against a generic hierarchy class."""

        TorsionLibraryInfo = self.TorsionLibraryInfo

        HierarchyClassElementNode = TorsionAlertsUtil.GetGenericHierarchyClassElementNode(TorsionLibraryInfo)
        if HierarchyClassElementNode is None:
            return (False, None)

        TorsionAlertsUtil.TrackHierarchyClassElementNode(TorsionLibraryInfo, HierarchyClassElementNode)

        #  Match hierarchy subclasses before matching torsion rules...
        MatchStatus, MatchInfo = self._MatchRotatableBondAgainstGenericHierarchySubClasses(Mol, RotBondAtomIndices, HierarchyClassElementNode)

        if not MatchStatus:
            MatchStatus, MatchInfo = self._MatchRotatableBondAgainstGenericHierarchyTorsionRules(Mol, RotBondAtomIndices, HierarchyClassElementNode)

        TorsionAlertsUtil.RemoveLastHierarchyClassElementNodeFromTracking(TorsionLibraryInfo)

        return (MatchStatus, MatchInfo)


    def _MatchRotatableBondAgainstGenericHierarchySubClasses(self, Mol, RotBondAtomIndices, HierarchyClassElementNode):
        """Match rotatable bond againat generic hierarchy subclasses."""

        for ElementChildNode in HierarchyClassElementNode:
            if ElementChildNode.tag != "hierarchySubClass":
                continue

            SubClassMatchStatus = self._ProcessHierarchySubClassElementForRotatableBondMatch(Mol, RotBondAtomIndices, ElementChildNode)

            if SubClassMatchStatus:
                MatchStatus, MatchInfo = self._ProcessElementForRotatableBondMatch(Mol, RotBondAtomIndices, ElementChildNode)

                if MatchStatus:
                    return (MatchStatus, MatchInfo)

        return(False, None)


    def _MatchRotatableBondAgainstGenericHierarchyTorsionRules(self, Mol, RotBondAtomIndices, HierarchyClassElementNode):
        """Match rotatable bond againat torsion rules generic hierarchy class."""

        for ElementChildNode in HierarchyClassElementNode:
            if ElementChildNode.tag != "torsionRule":
                continue

            MatchStatus, MatchInfo = self._ProcessTorsionRuleElementForRotatableBondMatch(Mol, RotBondAtomIndices, ElementChildNode)

            if MatchStatus:
                return (MatchStatus, MatchInfo)

        return(False, None)


    def _ProcessElementForRotatableBondMatch(self, Mol, RotBondAtomIndices, ElementNode):
        """Process element node to recursively match rotatable bond against hierarchy
        subclasses and torsion rules."""

        TorsionLibraryInfo = self.TorsionLibraryInfo

        for ElementChildNode in ElementNode:
            if ElementChildNode.tag == "hierarchySubClass":
                SubClassMatchStatus = self._ProcessHierarchySubClassElementForRotatableBondMatch(Mol, RotBondAtomIndices, ElementChildNode)

                if SubClassMatchStatus:
                    TorsionAlertsUtil.TrackHierarchySubClassElementNode(TorsionLibraryInfo, ElementChildNode)

                    MatchStatus, MatchInfo = self._ProcessElementForRotatableBondMatch(Mol, RotBondAtomIndices, ElementChildNode)
                    if MatchStatus:
                        TorsionAlertsUtil.RemoveLastHierarchySubClassElementNodeFromTracking(TorsionLibraryInfo)
                        return (MatchStatus, MatchInfo)

                    TorsionAlertsUtil.RemoveLastHierarchySubClassElementNodeFromTracking(TorsionLibraryInfo)

            elif ElementChildNode.tag == "torsionRule":
                MatchStatus, MatchInfo = self._ProcessTorsionRuleElementForRotatableBondMatch(Mol, RotBondAtomIndices, ElementChildNode)

                if MatchStatus:
                    return (MatchStatus, MatchInfo)

        return (False, None)


    def _ProcessHierarchySubClassElementForRotatableBondMatch(self, Mol, RotBondAtomIndices, ElementNode):
        """Process hierarchy subclass element to match rotatable bond."""

        # Setup subclass SMARTS pattern mol...
        SubClassPatternMol = TorsionAlertsUtil.SetupHierarchySubClassElementPatternMol(self.TorsionLibraryInfo, ElementNode)
        if SubClassPatternMol is None:
            return False

        # Match SMARTS pattern...
        SubClassPatternMatches = TorsionAlertsUtil.FilterSubstructureMatchesByAtomMapNumbers(Mol, SubClassPatternMol, Mol.GetSubstructMatches(SubClassPatternMol, useChirality = False))
        if len(SubClassPatternMatches) == 0:
            return False

        # Match rotatable bond indices...
        RotBondAtomIndex1, RotBondAtomIndex2 = RotBondAtomIndices
        MatchStatus = False
        for SubClassPatternMatch in SubClassPatternMatches:
            if len(SubClassPatternMatch) == 2:
                # Matched to pattern containing map atom numbers ":2" and ":3"...
                CentralAtomsIndex1, CentralAtomsIndex2 = SubClassPatternMatch
            elif len(SubClassPatternMatch) == 4:
                # Matched to pattern containing map atom numbers ":1", ":2", ":3" and ":4"...
                CentralAtomsIndex1 = SubClassPatternMatch[1]
                CentralAtomsIndex2 = SubClassPatternMatch[2]
            elif len(SubClassPatternMatch) == 3:
                SubClassSMARTSPattern = ElementNode.get("smarts")
                if TorsionAlertsUtil.DoesSMARTSContainsMappedAtoms(SubClassSMARTSPattern, [":2", ":3", ":4"]):
                    # Matched to pattern containing map atom numbers ":2", ":3" and ":4"...
                    CentralAtomsIndex1 = SubClassPatternMatch[0]
                    CentralAtomsIndex2 = SubClassPatternMatch[1]
                else:
                    # Matched to pattern containing map atom numbers ":1", ":2" and ":3"...
                    CentralAtomsIndex1 = SubClassPatternMatch[1]
                    CentralAtomsIndex2 = SubClassPatternMatch[2]
            else:
                continue

            if CentralAtomsIndex1 != CentralAtomsIndex2:
                if ((CentralAtomsIndex1 == RotBondAtomIndex1 and CentralAtomsIndex2 == RotBondAtomIndex2) or (CentralAtomsIndex1 == RotBondAtomIndex2 and CentralAtomsIndex2 == RotBondAtomIndex1)):
                    MatchStatus = True
                    break

        return (MatchStatus)


    def _ProcessTorsionRuleElementForRotatableBondMatch(self, Mol, RotBondAtomIndices, ElementNode):
        """Process torsion rule element to match rotatable bond."""

        TorsionLibraryInfo = self.TorsionLibraryInfo

        #  Retrieve torsion matched to rotatable bond...
        TorsionAtomIndices, TorsionAngle = self._MatchTorsionRuleToRotatableBond(Mol, RotBondAtomIndices, ElementNode)
        if TorsionAtomIndices is None:
            return (False, None)

        # Setup torsion angles info for matched torsion rule...
        TorsionAnglesInfo = TorsionAlertsUtil.SetupTorsionRuleAnglesInfo(TorsionLibraryInfo, ElementNode)
        if TorsionAnglesInfo is None:
            return (False, None)

        #   Setup torsion alert type and angle violation...
        AlertType, TorsionAngleViolation = self._SetupTorsionAlertTypeForRotatableBond(TorsionAnglesInfo, TorsionAngle)

        # Setup hierarchy class and subclass names...
        HierarchyClassName, HierarchySubClassName = TorsionAlertsUtil.SetupHierarchyClassAndSubClassNamesForRotatableBond(TorsionLibraryInfo)

        # Setup rule node ID...
        TorsionRuleNodeID = ElementNode.get("NodeID")

        # Setup SMARTS...
        TorsionRuleSMARTS = ElementNode.get("smarts")
        if " " in TorsionRuleSMARTS:
            TorsionRuleSMARTS = TorsionRuleSMARTS.replace(" ", "")

        # Setup torsion peaks and tolerances...
        TorsionRulePeaks = TorsionAnglesInfo["ValuesList"]
        TorsionRuleTolerances1 = TorsionAnglesInfo["Tolerances1List"]
        TorsionRuleTolerances2 = TorsionAnglesInfo["Tolerances2List"]

        MatchInfo = [AlertType, TorsionAtomIndices, TorsionAngle, TorsionAngleViolation, HierarchyClassName, HierarchySubClassName, TorsionRuleNodeID, TorsionRulePeaks, TorsionRuleTolerances1, TorsionRuleTolerances2, TorsionRuleSMARTS]

        # Setup match status...
        MatchStatus = True

        return (MatchStatus, MatchInfo)


    def _MatchTorsionRuleToRotatableBond(self, Mol, RotBondAtomIndices, ElementNode):
        """Retrieve matched torsion for torsion rule matched to rotatable bond."""

        # Get torsion matches...
        TorsionMatches = self._GetMatchesForTorsionRule(Mol, ElementNode)
        if TorsionMatches is None or len(TorsionMatches) == 0:
            return (None, None)

        # Identify the first torsion match corresponding to central atoms in RotBondAtomIndices...
        RotBondAtomIndex1, RotBondAtomIndex2 = RotBondAtomIndices
        for TorsionMatch in TorsionMatches:
            CentralAtomIndex1 = TorsionMatch[1]
            CentralAtomIndex2 = TorsionMatch[2]

            if ((CentralAtomIndex1 == RotBondAtomIndex1 and CentralAtomIndex2 == RotBondAtomIndex2) or (CentralAtomIndex1 == RotBondAtomIndex2 and CentralAtomIndex2 == RotBondAtomIndex1)):
                TorsionAngle = self._CalculateTorsionAngle(Mol, TorsionMatch)

                return (TorsionMatch, TorsionAngle)

        return (None, None)


    def _CalculateTorsionAngle(self, Mol, TorsionMatch):
        """Calculate torsion angle."""

        if type(TorsionMatch[3]) is list:
            return self._CalculateTorsionAngleUsingNitrogenLonePairPosition(Mol, TorsionMatch)

        # Calculate torsion angle using torsion atom indices..
        MolConf = Mol.GetConformer(0)
        TorsionAngle = rdMolTransforms.GetDihedralDeg(MolConf, TorsionMatch[0], TorsionMatch[1], TorsionMatch[2], TorsionMatch[3])
        TorsionAngle = round(TorsionAngle, 2)

        return TorsionAngle


    def _CalculateTorsionAngleUsingNitrogenLonePairPosition(self, Mol, TorsionMatch):
        """Calculate torsion angle using nitrogen lone pair positon."""

        # Setup a carbon atom as position holder for lone pair position...
        TmpMol = Chem.RWMol(Mol)
        LonePairAtomIndex = TmpMol.AddAtom(Chem.Atom(6))

        TmpMolConf = TmpMol.GetConformer(0)
        TmpMolConf.SetAtomPosition(LonePairAtomIndex, TorsionMatch[3])

        TorsionAngle = rdMolTransforms.GetDihedralDeg(TmpMolConf, TorsionMatch[0], TorsionMatch[1], TorsionMatch[2], LonePairAtomIndex)
        TorsionAngle = round(TorsionAngle, 2)

        return TorsionAngle


    def _GetMatchesForTorsionRule(self, Mol, ElementNode):
        """Get matches for torsion rule."""

        # Match torsions...
        TorsionMatches = None
        if self._IsNitogenLonePairTorsionRule(ElementNode):
            TorsionMatches = self._GetSubstructureMatchesForNitrogenLonePairTorsionRule(Mol, ElementNode)
        else:
            TorsionMatches = self._GetSubstructureMatchesForTorsionRule(Mol, ElementNode)

        if TorsionMatches is None or len(TorsionMatches) == 0:
            return TorsionMatches

        # Filter torsion matches...
        FiltertedTorsionMatches = []
        for TorsionMatch in TorsionMatches:
            if len(TorsionMatch) != 4:
                continue

            # Ignore matches containing hydrogen atoms as first or last atom...
            if Mol.GetAtomWithIdx(TorsionMatch[0]).GetAtomicNum() == 1:
                continue
            if type(TorsionMatch[3]) is int:
                # May contains a list for type two nitrogen lone pair match...
                if Mol.GetAtomWithIdx(TorsionMatch[3]).GetAtomicNum() == 1:
                    continue
            FiltertedTorsionMatches.append(TorsionMatch)

        return FiltertedTorsionMatches


    def _GetSubstructureMatchesForTorsionRule(self, Mol, ElementNode):
        """Get substructure matches for a torsion rule."""

        # Setup torsion rule SMARTS pattern mol....
        TorsionRuleNodeID = ElementNode.get("NodeID")
        TorsionSMARTSPattern = ElementNode.get("smarts")
        TorsionPatternMol = TorsionAlertsUtil.SetupTorsionRuleElementPatternMol(self.TorsionLibraryInfo, ElementNode, TorsionRuleNodeID, TorsionSMARTSPattern)
        if TorsionPatternMol is None:
            return None

        # Match torsions...
        TorsionMatches = TorsionAlertsUtil.FilterSubstructureMatchesByAtomMapNumbers(Mol, TorsionPatternMol, Mol.GetSubstructMatches(TorsionPatternMol, useChirality = False))

        return TorsionMatches


    def _GetSubstructureMatchesForNitrogenLonePairTorsionRule(self, Mol, ElementNode):
        """Get substructure matches for a torsion rule containing N_lp."""

        if self._IsTypeOneNitogenLonePairTorsionRule(ElementNode):
            return self._GetSubstructureMatchesForTypeOneNitrogenLonePairTorsionRule(Mol, ElementNode)
        elif self._IsTypeTwoNitogenLonePairTorsionRule(ElementNode):
            return self._GetSubstructureMatchesForTypeTwoNitrogenLonePairTorsionRule(Mol, ElementNode)

        return None


    def _GetSubstructureMatchesForTypeOneNitrogenLonePairTorsionRule(self, Mol, ElementNode):
        """Get substructure matches for a torsion rule containing N_lp and four mapped atoms."""

        # For example:
        #    [CX4:1][CX4H2:2]!@[NX3;"N_lp":3][CX4:4]
        #    [C:1][CX4H2:2]!@[NX3;"N_lp":3][C:4]
        #    ... ... ...

        TorsionRuleNodeID = ElementNode.get("NodeID")
        TorsionPatternMol, LonePairMapNumber = self._SetupNitrogenLonePairTorsionRuleElementInfo(ElementNode, TorsionRuleNodeID)

        if TorsionPatternMol is None:
            return None

        if LonePairMapNumber is None:
            return None

        # Match torsions...
        TorsionMatches = TorsionAlertsUtil.FilterSubstructureMatchesByAtomMapNumbers(Mol, TorsionPatternMol, Mol.GetSubstructMatches(TorsionPatternMol, useChirality = False))

        # Filter matches...
        FiltertedTorsionMatches = []
        for TorsionMatch in TorsionMatches:
            if len(TorsionMatch) != 4:
                continue

            # Check for Nitogen atom at LonePairMapNumber...
            LonePairNitrogenAtom = Mol.GetAtomWithIdx(TorsionMatch[LonePairMapNumber - 1])
            if LonePairNitrogenAtom.GetSymbol() != "N":
                continue

            # Make sure LonePairNitrogenAtom is planar...
            #  test
            PlanarityStatus = self._IsLonePairNitrogenAtomPlanar(Mol, LonePairNitrogenAtom)
            if PlanarityStatus is None:
                continue

            if  not PlanarityStatus:
                continue

            FiltertedTorsionMatches.append(TorsionMatch)

        return FiltertedTorsionMatches


    def _GetSubstructureMatchesForTypeTwoNitrogenLonePairTorsionRule(self, Mol, ElementNode):
        """Get substructure matches for a torsion rule containing N_lp and three mapped atoms."""

        # For example:
        # [!#1:1][CX4:2]!@[NX3;"N_lp":3]
        # [!#1:1][$(S(=O)=O):2]!@["N_lp":3]@[Cr3]
        # [C:1][$(S(=O)=O):2]!@["N_lp":3]
        # [c:1][$(S(=O)=O):2]!@["N_lp":3]
        # [!#1:1][$(S(=O)=O):2]!@["N_lp":3]
        #    ... ... ...

        TorsionRuleNodeID = ElementNode.get("NodeID")
        TorsionPatternMol, LonePairMapNumber = self._SetupNitrogenLonePairTorsionRuleElementInfo(ElementNode, TorsionRuleNodeID)

        if TorsionPatternMol is None:
            return None

        if not self._IsValidTypeTwoNitrogenLonePairMapNumber(LonePairMapNumber):
            return None

        # Match torsions...
        TorsionMatches = TorsionAlertsUtil.FilterSubstructureMatchesByAtomMapNumbers(Mol, TorsionPatternMol, Mol.GetSubstructMatches(TorsionPatternMol, useChirality = False))

        # Filter matches...
        FiltertedTorsionMatches = []
        for TorsionMatch in TorsionMatches:
            if len(TorsionMatch) != 3:
                continue

            # Check for Nitogen atom at LonePairMapNumber...
            LonePairNitrogenAtom = Mol.GetAtomWithIdx(TorsionMatch[LonePairMapNumber - 1])
            if LonePairNitrogenAtom.GetSymbol() != "N":
                continue

            # Make sure LonePairNitrogenAtom is not planar...
            PlanarityStatus = self._IsLonePairNitrogenAtomPlanar(Mol, LonePairNitrogenAtom)

            if PlanarityStatus is None:
                continue

            if  PlanarityStatus:
                continue

            # Calculate lone pair coordinates for a non-planar nitrogen...
            LonePairPosition = self._CalculateLonePairCoordinatesForNitrogenAtom(Mol, LonePairNitrogenAtom)
            if LonePairPosition is None:
                continue

            # Append lone pair coodinate list to list of torsion match containing atom indices...
            TorsionMatch.append(LonePairPosition)

            # Track torsion matches...
            FiltertedTorsionMatches.append(TorsionMatch)

        return FiltertedTorsionMatches


    def _SetupNitrogenLonePairTorsionRuleElementInfo(self, ElementNode, TorsionRuleNodeID):
        """Setup pattern molecule and lone pair map number for type one and type
        two nitrogen lone pair rules."""

        TorsionLibraryInfo = self.TorsionLibraryInfo
        TorsionPatternMol, LonePairMapNumber = [None] * 2

        if TorsionRuleNodeID in TorsionLibraryInfo["DataCache"]["TorsionRulePatternMol"]:
            TorsionPatternMol = TorsionLibraryInfo["DataCache"]["TorsionRulePatternMol"][TorsionRuleNodeID]
            LonePairMapNumber = TorsionLibraryInfo["DataCache"]["TorsionRuleLonePairMapNumber"][TorsionRuleNodeID]
        else:
            # Setup torsion pattern...
            TorsionSMARTSPattern = ElementNode.get("smarts")
            TorsionSMARTSPattern, LonePairMapNumber = self._ProcessSMARTSForNitrogenLonePairTorsionRule(TorsionSMARTSPattern)

            # Setup torsion pattern mol...
            TorsionPatternMol = Chem.MolFromSmarts(TorsionSMARTSPattern)
            if TorsionPatternMol is None:
                print("Warning: Ignoring torsion rule element containing invalid map atoms numbers in SMARTS pattern %s" % TorsionSMARTSPattern)

            # Cache data...
            TorsionLibraryInfo["DataCache"]["TorsionRulePatternMol"][TorsionRuleNodeID] = TorsionPatternMol
            TorsionLibraryInfo["DataCache"]["TorsionRuleLonePairMapNumber"][TorsionRuleNodeID] = LonePairMapNumber

        return (TorsionPatternMol, LonePairMapNumber)


    def _IsLonePairNitrogenAtomPlanar(self, Mol, NitrogenAtom):
        """Check for the planarity of nitrogen atom and its three neighbors."""

        AllowHydrogenNbrs = self.NitrogenLonePairAllowHydrogenNbrs
        Tolerance = self.NitrogenLonePairPlanarityTolerance

        # Get neighbors...
        if AllowHydrogenNbrs:
            AtomNeighbors = NitrogenAtom.GetNeighbors()
        else:
            AtomNeighbors = TorsionAlertsUtil.GetHeavyAtomNeighbors(NitrogenAtom)

        if len(AtomNeighbors) != 3:
            return None

        # Setup atom positions...
        AtomPositions = []
        MolAtomsPositions = TorsionAlertsUtil.GetAtomPositions(Mol)

        # Neighbor positions...
        for AtomNbr in AtomNeighbors:
            AtomNbrIndex = AtomNbr.GetIdx()
            AtomPositions.append(MolAtomsPositions[AtomNbrIndex])

        # Nitrogen position...
        NitrogenAtomIndex = NitrogenAtom.GetIdx()
        AtomPositions.append(MolAtomsPositions[NitrogenAtomIndex])

        Status =  self._AreFourPointsCoplanar(AtomPositions[0], AtomPositions[1], AtomPositions[2], AtomPositions[3], Tolerance)

        return Status


    def _AreFourPointsCoplanar(self, Point1, Point2, Point3, Point4, Tolerance = 1.0):
        """Check whether four points are coplanar with in the threshold of 1 degree."""

        # Setup  normalized direction vectors...
        VectorP2P1 = self._NormalizeVector(np.subtract(Point2, Point1))
        VectorP3P1 = self._NormalizeVector(np.subtract(Point3, Point1))
        VectorP1P4 = self._NormalizeVector(np.subtract(Point1, Point4))

        # Calculate angle between VectorP1P4 and normal to vectors VectorP2P1 and VectorP3P1...
        PlaneP1P2P3Normal = self._NormalizeVector(np.cross(VectorP2P1, VectorP3P1))
        PlanarityAngle = np.arccos(np.clip(np.dot(PlaneP1P2P3Normal, VectorP1P4), -1.0, 1.0))

        Status = math.isclose(PlanarityAngle, math.radians(90), abs_tol=math.radians(Tolerance))

        return Status


    def _NormalizeVector(self, Vector):
        """Normalize vector."""

        Norm = np.linalg.norm(Vector)

        return Vector if math.isclose(Norm, 0.0, abs_tol = 1e-08) else Vector/Norm


    def _CalculateLonePairCoordinatesForNitrogenAtom(self, Mol, NitrogenAtom):
        """Calculate approximate lone pair coordinates for non-plannar nitrogen atom."""

        AllowHydrogenNbrs = self.NitrogenLonePairAllowHydrogenNbrs

        # Get neighbors...
        if AllowHydrogenNbrs:
            AtomNeighbors = NitrogenAtom.GetNeighbors()
        else:
            AtomNeighbors = TorsionAlertsUtil.GetHeavyAtomNeighbors(NitrogenAtom)

        if len(AtomNeighbors) != 3:
            return None

        # Setup positions for nitrogen and its neghbors...
        MolAtomsPositions = TorsionAlertsUtil.GetAtomPositions(Mol)

        NitrogenPosition = MolAtomsPositions[NitrogenAtom.GetIdx()]
        NbrPositions = []
        for AtomNbr in AtomNeighbors:
            NbrPositions.append(MolAtomsPositions[AtomNbr.GetIdx()])
        Nbr1Position, Nbr2Position, Nbr3Position = NbrPositions

        # Setup  normalized direction vectors...
        VectorP2P1 = self._NormalizeVector(np.subtract(Nbr2Position, Nbr1Position))
        VectorP3P1 = self._NormalizeVector(np.subtract(Nbr3Position, Nbr1Position))
        VectorP1P4 = self._NormalizeVector(np.subtract(Nbr1Position, NitrogenPosition))

        # Calculate angle between VectorP1P4 and normal to vectors VectorP2P1 and VectorP3P1...
        PlaneP1P2P3Normal = self._NormalizeVector(np.cross(VectorP2P1, VectorP3P1))
        PlanarityAngle = np.arccos(np.clip(np.dot(PlaneP1P2P3Normal, VectorP1P4), -1.0, 1.0))

        # Check for reversing the direction of the normal...
        if PlanarityAngle < math.radians(90):
            PlaneP1P2P3Normal = PlaneP1P2P3Normal * -1

        # Add normal to nitrogen cooridnates for the approximate coordinates of the
        # one pair. The exact VSEPR coordinates of the lone pair are not necessary to
        # calculate the torsion angle...
        LonePairPosition = NitrogenPosition + PlaneP1P2P3Normal

        return list(LonePairPosition)


    def _ProcessSMARTSForNitrogenLonePairTorsionRule(self, SMARTSPattern):
        """Process SMARTS pattern for a torion rule containing N_lp."""

        LonePairMapNumber = self._GetNitrogenLonePairMapNumber(SMARTSPattern)

        # Remove double quotes around N_lp..
        SMARTSPattern = re.sub("\"N_lp\"", "N_lp", SMARTSPattern, re.I)

        # Remove N_lp specification from SMARTS pattern for torsion rule...
        if re.search("\[N_lp", SMARTSPattern, re.I):
            # Handle missing NX3...
            SMARTSPattern = re.sub("\[N_lp", "[NX3", SMARTSPattern)
        else:
            SMARTSPattern = re.sub(";N_lp", "", SMARTSPattern)

        return (SMARTSPattern, LonePairMapNumber)


    def _GetNitrogenLonePairMapNumber(self, SMARTSPattern):
        """Get atom map number for nitrogen involved in N_lp."""

        LonePairMapNumber = None

        SMARTSPattern = re.sub("\"N_lp\"", "N_lp", SMARTSPattern, re.I)
        MatchedMappedAtoms = re.findall("N_lp:[0-9]", SMARTSPattern, re.I)

        if len(MatchedMappedAtoms) == 1:
            LonePairMapNumber = int(re.sub("N_lp:", "", MatchedMappedAtoms[0]))

        return LonePairMapNumber


    def _IsNitogenLonePairTorsionRule(self, ElementNode):
        """Check for the presence of N_lp in SMARTS pattern for a torsion rule."""

        if "N_lp" not  in ElementNode.get("smarts"):
            return False

        LonePairMatches = re.findall("N_lp", ElementNode.get("smarts"), re.I)

        return True if len(LonePairMatches) == 1 else False


    def _IsTypeOneNitogenLonePairTorsionRule(self, ElementNode):
        """Check for the presence four mapped atoms in a SMARTS pattern containing
        N_lp for a torsion rule."""

        # For example:
        #    [CX4:1][CX4H2:2]!@[NX3;"N_lp":3][CX4:4]
        #    [C:1][CX4H2:2]!@[NX3;"N_lp":3][C:4]
        #    ... ... ...

        MatchedMappedAtoms = re.findall(":[0-9]", ElementNode.get("smarts"), re.I)

        return True if len(MatchedMappedAtoms) == 4 else False


    def _IsTypeTwoNitogenLonePairTorsionRule(self, ElementNode):
        """Check for the presence three mapped atoms in a SMARTS pattern containing
        N_lp for a torsion rule."""

        # For example:
        # [!#1:1][CX4:2]!@[NX3;"N_lp":3]
        # [!#1:1][$(S(=O)=O):2]!@["N_lp":3]@[Cr3]
        # [C:1][$(S(=O)=O):2]!@["N_lp":3]
        # [c:1][$(S(=O)=O):2]!@["N_lp":3]
        # [!#1:1][$(S(=O)=O):2]!@["N_lp":3]
        #

        MatchedMappedAtoms = re.findall(":[0-9]", ElementNode.get("smarts"), re.I)

        return True if len(MatchedMappedAtoms) == 3 else False


    def _IsValidTypeTwoNitogenLonePairTorsionRule(self, ElementNode):
        """Validate atom map number for nitrogen involved in N_lp for type two nitrogen
        lone pair torsion rule."""

        LonePairMapNumber = self._GetNitrogenLonePairMapNumber(ElementNode.get("smarts"))

        return self._IsValidTypeTwoNitrogenLonePairMapNumber(LonePairMapNumber)


    def _IsValidTypeTwoNitrogenLonePairMapNumber(self, LonePairMapNumber):
        """Check that  the atom map number is 3."""

        return True if LonePairMapNumber is not None and LonePairMapNumber == 3 else False


    def _SetupTorsionAlertTypeForRotatableBond(self, TorsionAnglesInfo, TorsionAngle):
        """Setup torsion alert type and angle violation for a rotatable bond."""

        TorsionCategory, TorsionAngleViolation = [None, None]

        for ID in TorsionAnglesInfo["IDs"]:
            if self._IsTorsionAngleInWithinTolerance(TorsionAngle, TorsionAnglesInfo["Value"][ID], TorsionAnglesInfo["Tolerance1"][ID]):
                TorsionCategory = "Green"
                TorsionAngleViolation = 0.0
                break

            if self._IsTorsionAngleInWithinTolerance(TorsionAngle, TorsionAnglesInfo["Value"][ID], TorsionAnglesInfo["Tolerance2"][ID]):
                TorsionCategory = "Orange"
                TorsionAngleViolation = self._CalculateTorsionAngleViolation(TorsionAngle, TorsionAnglesInfo["ValuesIn360RangeList"], TorsionAnglesInfo["Tolerances1List"])
                break

        if TorsionCategory is None:
            TorsionCategory = "Red"
            TorsionAngleViolation = self._CalculateTorsionAngleViolation(TorsionAngle, TorsionAnglesInfo["ValuesIn360RangeList"], TorsionAnglesInfo["Tolerances2List"])

        return (TorsionCategory, TorsionAngleViolation)


    def _IsTorsionAngleInWithinTolerance(self, TorsionAngle, TorsionPeak, TorsionTolerance):
        """Check torsion angle against torsion tolerance."""

        TorsionAngleDiff = TorsionAlertsUtil.CalculateTorsionAngleDifference(TorsionPeak, TorsionAngle)

        return True if (abs(TorsionAngleDiff) <= TorsionTolerance) else False


    def _CalculateTorsionAngleViolation(self, TorsionAngle, TorsionPeaks, TorsionTolerances):
        """Calculate torsion angle violation."""

        TorsionAngleViolation = None

        # Map angle to 0 to 360 range. TorsionPeaks values must be in this range...
        if TorsionAngle < 0:
            TorsionAngle = TorsionAngle + 360

        # Identify the closest torsion peak index...
        if len(TorsionPeaks) == 1:
            NearestPeakIndex = 0
        else:
            NearestPeakIndex = min(range(len(TorsionPeaks)), key=lambda Index: abs(TorsionPeaks[Index] - TorsionAngle))

        # Calculate torsion angle violation from the nearest peak and its tolerance value...
        TorsionAngleDiff = TorsionAlertsUtil.CalculateTorsionAngleDifference(TorsionPeaks[NearestPeakIndex], TorsionAngle)
        TorsionAngleViolation = abs(abs(TorsionAngleDiff) - TorsionTolerances[NearestPeakIndex])

        return TorsionAngleViolation
