#
# File: TorsionStrainEnergyAlerts.py
# Author: Manish Sud <msud@san.rr.com>
#
# Collaborator: Pat Walters
#
# Copyright (C) 2024 Manish Sud. All rights reserved.
#
# This module uses the torsion strain energy library developed by Gu, S.;
# Smith, M. S.; Yang, Y.; Irwin, J. J.; Shoichet, B. K. [ Ref 153 ].
#
# The torsion strain enegy library is based on the Torsion Library jointly
# developed by the University of Hamburg, Center for Bioinformatics,
# Hamburg, Germany and F. Hoffmann-La-Roche Ltd., Basel, Switzerland.
#
# This file is part of MayaChemTools.
#
# MayaChemTools is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3 of the License, or (at your option) any
# later version.
#
# MayaChemTools is distributed in the hope that it will be useful, but without
# any warranty; without even the implied warranty of merchantability of fitness
# for a particular purpose.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with MayaChemTools; if not, see <http://www.gnu.org/licenses/> or
# write to the Free Software Foundation Inc., 59 Temple Place, Suite 330,
# Boston, MA, 02111-1307, USA.
#

import os
import re
import math

from rdkit import Chem
from rdkit.Chem import rdMolTransforms

from . import TorsionAlertsUtil

class TorsionStrainEnergyAlerts():
    def __init__(self, AlertsMode = "TotalEnergy", TotalEnergyCutoff = 6.0, MaxSingleEnergyCutoff = 1.8, RotBondsSMARTSMode = "SemiStrict", RotBondsSMARTSPattern = None, TorsionLibraryFilePath = "auto", AlertTorsionsNotObserved = False):
        """Identify strained molecules based on torsion strain energy library [ Ref 153 ]
        alerts by matching rotatable bonds against SMARTS patterns specified for
        torsion rules in a torsion energy library file. The molecules must have 3D coordinates.
        The default torsion strain energy library file, TorsionStrainEnergyLibrary.xml, is
        ia available in the directory containing this file.
        
        The data in torsion strain energy library file is organized in a hierarchical
        manner. It consists of one generic class and six specific classes at the highest
        level. Each class contains multiple subclasses corresponding to named functional
        groups or substructure patterns. The subclasses consist of torsion rules sorted
        from specific to generic torsion patterns. The torsion rule, in turn, contains a
        list of peak values for torsion angles and two tolerance values. A pair of tolerance
        values define torsion bins around a torsion peak value.
        
        A strain energy calculation method, 'exact' or 'approximate' [ Ref 153 ], is 
        associated with each torsion rule for calculating torsion strain energy. The 'exact'
        stain energy calculation relies on the energy bins available under the energy histogram
        consisting of 36 bins covering angles from -180 to 180. The width of each bin is 10
        degree. The energy bins are are defined at the right end points. The first and the
        last energy bins correspond to -170 and 180 respectively. The torsion angle is mapped
        to a energy bin. An angle offset is calculated for the torsion angle from the the right
        end point angle of the bin. The strain energy is estimated for the angle offset based
        on the energy difference between the current and previous bins. The torsion strain
        energy, in terms of torsion energy units (TEUs), corresponds to the sum of bin strain
        energy and the angle offset strain energy.
            
            Energy = BinEnergyDiff/10.0 * BinAngleOffset + BinEnergy[BinNum]
            
            Where:
            
            BinEnergyDiff = BinEnergy[BinNum] - BinEnergy[PreviousBinNum]
            BinAngleOffset = TorsionAngle - BinAngleRightSide
            
        The 'approximate' strain energy calculation relies on the angle difference between a
        torsion angle and the torsion peaks observed for the torsion rules in the torsion
        energy library. The torsion angle is matched to a torsion peak based on the value of
        torsion angle difference. It must be less than or equal to the value for the second
        tolerance 'tolerance2'. Otherwise, the torsion angle is not observed in the torsion
        energy library and a value of 'NA' is assigned for torsion energy along with the lower
        and upper bounds on energy at 95% confidence interval. The 'approximate' torsion
        energy (TEUs) for observed torsion angle is calculated using the following formula:
            
            Energy = beta_1 * (AngleDiff ** 2) + beta_2 * (AngleDiff ** 4)
            
        The coefficients 'beta_1' and 'beta_2' are available for the observed angles in 
        the torsion strain energy library. The 'AngleDiff' is the difference between the
        torsion angle and the matched torsion peak.
        
        For example:
             
            <library>
                <hierarchyClass id1="G" id2="G" name="GG">
                ...
                </hierarchyClass>
                <hierarchyClass id1="C" id2="O" name="CO">
                    <hierarchySubClass name="Ester bond I" smarts="O=[C:2][O:3]">
                        <torsionRule method="exact" smarts=
                            "[O:1]=[C:2]!@[O:3]~[CH0:4]">
                            <angleList>
                                <angle score="56.52" tolerance1="20.00"
                                tolerance2="25.00" value="0.0"/>
                            </angleList>
                            <histogram>
                                <bin count="1"/>
                                ...
                            </histogram>
                            <histogram_shifted>
                                <bin count="0"/>
                                ...
                            </histogram_shifted>
                            <histogram_converted>
                                <bin energy="4.67... lower="2.14..." upper="Inf"/>
                                ...
                                <bin energy="1.86..." lower="1.58..." upper="2.40..."/>
                                ...
                               </histogram_converted>
                        </torsionRule>
                        <torsionRule method="approximate" smarts=
                            "[cH0:1][c:2]([cH0])!@[O:3][p:4]">
                            <angleList>
                            <angle beta_1="0.002..." beta_2="-7.843...e-07"
                                score="27.14" theta_0="-90.0" tolerance1="30.00"
                                tolerance2="45.00" value="-90.0"/>
                            ...
                            </angleList>
                            <histogram>
                                <bin count="0"/>
                                 ...
                            </histogram>
                            <histogram_shifted>
                                <bin count="0"/>
                                ...
                            </histogram_shifted>
                        </torsionRule>
                    ...
                 ...
                </hierarchyClass>
                 <hierarchyClass id1="N" id2="C" name="NC">
                 ...
                </hierarchyClass>
                <hierarchyClass id1="S" id2="N" name="SN">
                ...
                </hierarchyClass>
                <hierarchyClass id1="C" id2="S" name="CS">
                ...
                </hierarchyClass>
                <hierarchyClass id1="C" id2="C" name="CC">
                ...
                </hierarchyClass>
                <hierarchyClass id1="S" id2="S" name="SS">
                 ...
                </hierarchyClass>
            </library>
            
        The rotatable bonds in a 3D molecule are identified using a default SMARTS pattern.
        A custom SMARTS pattern may be optionally specified to detect rotatable bonds.
        Each rotatable bond is matched to a torsion rule in the torsion strain energy library.
        The strain energy is calculated for each rotatable bond using the calculation
        method, 'exact' or 'approximate', associated with the matched torsion rule.
        
        The total strain energy (TEUs) of a molecule corresponds to the sum of  'exact' and
        'approximate' strain energies calculated for all matched rotatable bonds in the
        molecule. The total strain energy is set to 'NA' for molecules containing a 'approximate'
        energy estimate for a torsion angle not observed in the torsion energy library. In
        addition, the lower and upper bounds on energy at 95% confidence interval are
        set to 'NA'.

        Arguments:
            AlertsMode (str): Torsion strain energy library alert types to use
                for issuing alerts. Possible values: TotalEnergy,
                MaxSingleEnergy, or TotalOrMaxSingleEnergy.
            TotalEnergyCutoff (float): Total strain strain energy (TEUs) cutoff.
            MaxSingleEnergyCutoff (float): Maximum single strain energy (TEUs)
                cutoff.
            RotBondsSMARTSMode (str): SMARTS pattern to use for identifying
                rotatable bonds in a molecule. Possible values: NonStrict,
                SemiStrict, Strict or Specify.
            RotBondsSMARTSPattern (str):SMARTS pattern for identifying rotatable
                bonds. This paramater is only valid for 'Specify' value
                'RotBondsSMARTSMode'.
            TorsionLibraryFilePath (str): A XML file name containing data for
                torsion starin energy library.
            AlertTorsionsNotObserved (bool): Issue alerts about molecules
                containing torsion angles not observed in torsion strain energy
                library.

        Returns:
            object: An instantiated class object.

        Examples:
            
            StrainEnergyAlertsHandle = TorsionStrainEnergyAlerts()

            StrainEnergyAlertsHandle = TorsionStrainEnergyAlerts(AlertsMode = "TotalEnergy",
                TotalEnergyCutoff = 6.0, MaxSingleEnergyCutoff = 1.8, RotBondsSMARTSMode = "SemiStrict",
                RotBondsSMARTSPattern = None, TorsionLibraryFilePath = "auto", AlertTorsionsNotObserved = False)

        Notes:
            The following sections provide additional details for the parameters.
            
            AlertsMode: Torsion strain energy library alert types to use for issuing 
            alerts about molecules containing rotatable bonds based on the calculated
            values for the total torsion strain energy of a molecule and  the maximum
            single strain energy of a rotatable bond in a molecule.
            
            Possible values: TotalEnergy, MaxSingleEnergy, or TotalOrMaxSingleEnergy
            
            The strain energy cutoff values in terms of torsion energy units (TEUs) are
            used to filter molecules as shown below:
                
                AlertsMode                AlertsEnergyCutoffs (TEUs)
                
                TotalEnergy               >= TotalEnergyCutoff
                
                MaxSingleEnergy           >= MaxSingleEnergyCutoff
                
                TotalOrMaxSingleEnergy    >= TotalEnergyCutoff
                                          or >= MaxSingleEnergyCutoff
            
            TotalEnergyCutoff: Total strain strain energy (TEUs) cutoff [ Ref 153 ] for
            issuing alerts based on total strain energy for all rotatable bonds in a
            molecule. This option is used during 'TotalEnergy' or 'TotalOrMaxSingleEnergy'
            values of 'AlertsMode' parameter.
            
            The total strain energy must be greater than or equal to the specified
            cutoff value to identify a molecule containing strained torsions.
            
            MaxSingleEnergyCutoff: Maximum single strain energy (TEUs) cutoff [ Ref 153 ]
            for issuing alerts based on the maximum value of a single strain energy of a
            rotatable bond in  a molecule. This option is used during 'MaxSingleEnergy' or
            'TotalOrMaxSingleEnergy' values of 'AlertsMode' parameter.
            
            The maximum single strain energy must be greater than or equal to the
            specified cutoff valueo to identify a molecule containing strained torsions.
            
            RotBondsSMARTSMode: SMARTS pattern to use for identifying rotatable bonds in
            a molecule for matching against torsion rules in the torsion library. Possible values:
            NonStrict, SemiStrict, Strict or Specify. The rotatable bond SMARTS matches
            are filtered to ensure that each atom in the rotatable bond is attached to
            at least two heavy atoms.
            
            The following SMARTS patterns are used to identify rotatable bonds for
            different modes:
                
                NonStrict: [!$(*#*)&!D1]-&!@[!$(*#*)&!D1]
                
                SemiStrict:
                [!$(*#*)&!D1&!$(C(F)(F)F)&!$(C(Cl)(Cl)Cl)&!$(C(Br)(Br)Br)
                &!$(C([CH3])([CH3])[CH3])]-!@[!$(*#*)&!D1&!$(C(F)(F)F)
                &!$(C(Cl)(Cl)Cl)&!$(C(Br)(Br)Br)&!$(C([CH3])([CH3])[CH3])]
                
                Strict:
                [!$(*#*)&!D1&!$(C(F)(F)F)&!$(C(Cl)(Cl)Cl)&!$(C(Br)(Br)Br)
                &!$(C([CH3])([CH3])[CH3])&!$([CD3](=[N,O,S])-!@[#7,O,S!D1])
                &!$([#7,O,S!D1]-!@[CD3]=[N,O,S])&!$([CD3](=[N+])-!@[#7!D1])
                &!$([#7!D1]-!@[CD3]=[N+])]-!@[!$(*#*)&!D1&!$(C(F)(F)F)
                &!$(C(Cl)(Cl)Cl)&!$(C(Br)(Br)Br)&!$(C([CH3])([CH3])[CH3])]
                
            The 'NonStrict' and 'Strict' SMARTS patterns are available in RDKit. The 
            'NonStrict' SMARTS pattern corresponds to original Daylight SMARTS
            specification for rotatable bonds. The 'SemiStrict' SMARTS pattern is 
            derived from 'Strict' SMARTS pattern.
            
            TorsionLibraryFilePath (str): Specify a XML file name containing data for
            torsion starin energy library hierarchy or use default file,
            TorsionEnergyLibrary.xml, available in the directory containing this file.
        
            AlertTorsionsNotObserved: Issue alerts abpout molecules con ntaining torsion
            angles not observed in torsion strain energy library. It's not possible to
            calculate torsion strain energies for these torsions during 'approximate'
            match to a specified torsion in the library.
            
            The 'approximate' strain energy calculation relies on the angle difference
            between a torsion angle and the torsion peaks observed for the torsion
            rules in the torsion energy library. The torsion angle is matched to a
            torsion peak based on the value of torsion angle difference. It must be
            less than or equal to the value for the second tolerance 'tolerance2'.
            Otherwise, the torsion angle is not observed in the torsion energy library
            and a value of 'NA' is assigned for torsion energy along with the lower and
            upper bounds on energy at 95% confidence interval.

        """

        self._ProcessTorsionStrainEnergyAlertsParameters(AlertsMode, TotalEnergyCutoff, MaxSingleEnergyCutoff, RotBondsSMARTSMode, RotBondsSMARTSPattern, TorsionLibraryFilePath, AlertTorsionsNotObserved)

        self._InitializeTorsionStrainEnergyAlerts()


    def _ProcessTorsionStrainEnergyAlertsParameters(self, AlertsMode, TotalEnergyCutoff, MaxSingleEnergyCutoff, RotBondsSMARTSMode, RotBondsSMARTSPattern, TorsionLibraryFilePath, AlertTorsionsNotObserved):
        """Process torsion strain energy alerts paramaters. """

        # Process AltertsMode paramater...
        self._ProcessAlertsModeParameter(AlertsMode)

        # Process TotalEnergyCutoff parameter...
        if TotalEnergyCutoff <= 0.0:
            raise ValueError("The value, %s, specified for TotalEnergyCutoff parameter is not valid. Supported value: > 0.0" % TotalEnergyCutoff)
        self.TotalEnergyCutoff = TotalEnergyCutoff

        #Process  MaxSingleEnergyCutoff parameter...
        if MaxSingleEnergyCutoff <= 0.0:
            raise ValueError("The value, %s, specified for MaxSingleEnergyCutoff parameter is not valid. Supported value: > 0.0" % MaxSingleEnergyCutoff)
        self.MaxSingleEnergyCutoff = MaxSingleEnergyCutoff

        # Process RotBondsSMARTSMode and RotBondsSMARTSPattern parameters...
        self._ProcessRotBondsParameters(RotBondsSMARTSMode, RotBondsSMARTSPattern)

        # Process TorsionLibraryFilePath parameter...
        self._ProcessTorsionLibraryFilePathParameter(TorsionLibraryFilePath)

        # AlertTorsionsNotObserved parameter...
        self.AlertTorsionsNotObserved = AlertTorsionsNotObserved


    def _InitializeTorsionStrainEnergyAlerts(self):
        """Initialize tosion strain energy alerts."""

        # Retrieve and setup torsion library info for matching rotatable bonds...
        TorsionLibraryInfo = {}

        TorsionLibElementTree = TorsionAlertsUtil.RetrieveTorsionLibraryInfo(self.TorsionLibraryFilePath)
        TorsionLibraryInfo["TorsionLibElementTree"] = TorsionLibElementTree

        TorsionAlertsUtil.SetupTorsionLibraryInfoForMatchingRotatableBonds(TorsionLibraryInfo)

        self.TorsionLibraryInfo = TorsionLibraryInfo


    def _ProcessAlertsModeParameter(self, AlertsMode):
        """Process AlertsMode parameter. """

        TotalEnergyMode, MaxSingleEnergyMode, TotalOrMaxSingleEnergyMode = [False] * 3
        if re.match("^TotalEnergy$", AlertsMode, re.I):
            TotalEnergyMode = True
        elif re.match("^MaxSingleEnergy$", AlertsMode, re.I):
            MaxSingleEnergyMode = True
        elif re.match("^TotalOrMaxSingleEnergy$", AlertsMode, re.I):
            TotalOrMaxSingleEnergyMode = True
        else:
            raise ValueError("Invalid value, %s, specified for AlertsMode parameter. Valid values: TotalEnergy, MaxSingleEnergy, or TotalOrMaxSingleEnergy" % (AlertsMode))

        self.AltersMode = AlertsMode
        self.TotalEnergyMode = TotalEnergyMode
        self.MaxSingleEnergyMode = MaxSingleEnergyMode
        self.TotalOrMaxSingleEnergyMode = TotalOrMaxSingleEnergyMode


    def _ProcessRotBondsParameters(self, RotBondsSMARTSMode, RotBondsSMARTSPattern):
        """Process  RotBondsSMARTSMode and RotBondsSMARTSPattern parameters. """

        if re.match("^NonStrict$", RotBondsSMARTSMode, re.I):
            RotBondsSMARTSPattern = "[!$(*#*)&!D1]-&!@[!$(*#*)&!D1]"
        elif re.match("^SemiStrict$", RotBondsSMARTSMode, re.I):
            RotBondsSMARTSPattern = "[!$(*#*)&!D1&!$(C(F)(F)F)&!$(C(Cl)(Cl)Cl)&!$(C(Br)(Br)Br)&!$(C([CH3])([CH3])[CH3])]-!@[!$(*#*)&!D1&!$(C(F)(F)F)&!$(C(Cl)(Cl)Cl)&!$(C(Br)(Br)Br)&!$(C([CH3])([CH3])[CH3])]"
        elif re.match("^Strict$", RotBondsSMARTSMode, re.I):
            RotBondsSMARTSPattern = "[!$(*#*)&!D1&!$(C(F)(F)F)&!$(C(Cl)(Cl)Cl)&!$(C(Br)(Br)Br)&!$(C([CH3])([CH3])[CH3])&!$([CD3](=[N,O,S])-!@[#7,O,S!D1])&!$([#7,O,S!D1]-!@[CD3]=[N,O,S])&!$([CD3](=[N+])-!@[#7!D1])&!$([#7!D1]-!@[CD3]=[N+])]-!@[!$(*#*)&!D1&!$(C(F)(F)F)&!$(C(Cl)(Cl)Cl)&!$(C(Br)(Br)Br)&!$(C([CH3])([CH3])[CH3])]"
        elif re.match("^Specify$", RotBondsSMARTSMode, re.I):
            if RotBondsSMARTSPattern is None:
                raise ValueError("The RotBondsSMARTSPattern parameter value must be specified during Specify value for RotBondsSMARTSMode parameter.")
            RotBondsSMARTSPattern = RotBondsSMARTSPattern.strip()
            if not len(RotBondsSMARTSPattern):
                raise ValueError("Empty value specified for  RotBondsSMARTSPattern parameter.")
        else:
            raise ValueError("Invalid value, %s, specified for RotBondsSMARTSMode parameter. Valid values: NonStrict, SemiStrict, Strict or Specify" % (RotBondsSMARTSMode))

        RotBondsPatternMol = Chem.MolFromSmarts(RotBondsSMARTSPattern)
        if RotBondsPatternMol is None:
            if re.match("Specify", RotBondsSMARTSMode, re.I):
                raise ValueError("Failed to create rotatable bonds pattern molecule. The rotatable bonds SMARTS pattern, \"%s\", specified using RotBondsSMARTSPattern parameter is not valid." % (RotBondsSMARTSPattern))
            else:
                raise ValueError("Failed to create rotatable bonds pattern molecule. The default rotatable bonds SMARTS pattern, \"%s\", used for %s value of RotBondsSMARTSMode parameter is not valid." % (RotBondsSMARTSPattern, RotBondsSMARTSMode))

        self.RotBondsSMARTSMode = RotBondsSMARTSMode
        self.RotBondsSMARTSPattern = RotBondsSMARTSPattern
        self.RotBondsPatternMol = RotBondsPatternMol


    def _ProcessTorsionLibraryFilePathParameter(self, TorsionLibraryFilePath):
        """ Process torsion library path."""

        # TorsionLibraryFilePath parameter...
        if re.match("^auto$", TorsionLibraryFilePath):
            TorsionLibraryFile = "TorsionStrainEnergyLibrary.xml"
            TorsionLibraryFilePath = os.path.join(os.path.dirname(os.path.abspath(__file__)), TorsionLibraryFile)
            if not os.path.isfile(TorsionLibraryFilePath):
                raise ValueError("The default torsion alerts library file %s doesn't exist." % TorsionLibraryFilePath)
        else:
            if not os.path.isfile(TorsionLibraryFilePath):
                raise ValueError("The file specified, %s, for parameter TorsionLibraryFilePath doesn't exist." % TorsionLibraryFilePath)
            TorsionLibraryFilePath = os.path.abspath(TorsionLibraryFilePath)

        self.TorsionLibraryFilePath = TorsionLibraryFilePath


    def GetTorsionLibraryFilePath(self):
        """Get torsion strain library file path.

        Arguments:
            None

        Returns:
            FilePath (str): Torsion strain library path.

        """

        return self.TorsionLibraryFilePath


    def ListTorsionLibraryInfo(self):
        """List torsion strain library information.

        Arguments:
            None

        Returns:
            None

        """

        return TorsionAlertsUtil.ListTorsionLibraryInfo(self.TorsionLibraryInfo["TorsionLibElementTree"])


    def IdentifyTorsionLibraryAlertsForRotatableBonds(self, Mol):
        """Identify torsion strain library alerts for a molecule by matching rotatable
        bonds against SMARTS patterns specified for torsion rules in torsion energy
        library file.
 
        Arguments:
            Mol (object): RDKit molecule object.

        Returns:
            bool: True - Molecule contains strained torsions; False - Molecule
                contains no strained torsions or rotatable bonds. 
            dict or None: Torsion alerts information regarding matching of
                rotatable bonds to torsion strain library.

        Examples:

            from TorsionAlerts.TorsionStrainEnergyAlerts import
                TorsionStrainEnergyAlerts
            
            StrainEnergyAlerts = TorsionStrainEnergyAlerts()
            AlertsStatus, AlertsInfo = StrainEnergyAlerts.
                IdentifyTorsionLibraryAlertsForRotatableBonds(RDKitMol)
            
            RotBondsAlertsStatus = AlertsInfo["RotBondsAlertsStatus"]
            TotalEnergy = AlertsInfo["TotalEnergy"]
            TotalEnergyLowerBound = AlertsInfo["TotalEnergyLowerBound"]
            TotalEnergyUpperBound = AlertsInfo["TotalEnergyUpperBound"]
            AnglesNotObservedCount = AlertsInfo["AnglesNotObservedCount"]
            MaxSingleEnergy = AlertsInfo["MaxSingleEnergy"])
            MaxSingleEnergyAlertsCount = AlertsInfo[
                "MaxSingleEnergyAlertsCount"]
            
            # List of rotatable bond IDs...
            RotatableBondIDs = AlertsInfo["IDs"]
            
            # Dictionaries containing information for rotatable bonds by using
            # bond ID as key...
            for ID in AlertsInfo["IDs"]:
                MatchStatus = AlertsInfo["MatchStatus"][ID]
                MaxSingleEnergyAlertStatus = AlertsInfo[
                    "MaxSingleEnergyAlertStatus"][ID]
                AtomIndices = AlertsInfo["AtomIndices"][ID]
                TorsionAtomIndices = AlertsInfo["TorsionAtomIndices"][ID]
                TorsionAngle = AlertsInfo["TorsionAngle"][ID]
                HierarchyClassName = AlertsInfo["HierarchyClassName"][ID]
                HierarchySubClassName = AlertsInfo["HierarchySubClassName"][ID]
                TorsionRuleNodeID = AlertsInfo["TorsionRuleNodeID"][ID]
                TorsionRuleSMARTS = AlertsInfo["TorsionRuleSMARTS"][ID]
                EnergyMethod = AlertsInfo["EnergyMethod"][ID]
                AngleNotObserved = AlertsInfo["AngleNotObserved"][ID]
                Energy = AlertsInfo["Energy"][ID]
                EnergyLowerBound = AlertsInfo["EnergyLowerBound"][ID]
                EnergyUpperBound = AlertsInfo["EnergyUpperBound"][ID]

        """

        # Identify rotatable bonds...
        RotBondsStatus, RotBondsInfo = TorsionAlertsUtil.IdentifyRotatableBondsForTorsionLibraryMatch(self.TorsionLibraryInfo, Mol, self.RotBondsPatternMol)

        if not RotBondsStatus:
            return (False, None)

        # Identify alerts for rotatable bonds...
        RotBondsAlertsStatus, RotBondsAlertsInfo = self._MatchRotatableBondsToTorsionLibrary(Mol, RotBondsInfo)

        return (RotBondsAlertsStatus, RotBondsAlertsInfo)


    def _MatchRotatableBondsToTorsionLibrary(self, Mol, RotBondsInfo):
        """Match rotatable bond to torsion library."""

        # Initialize...
        RotBondsAlertsInfo = self._InitializeRotatableBondsAlertsInfo()

        # Match rotatable bonds to torsion library...
        for ID in RotBondsInfo["IDs"]:
            AtomIndices = RotBondsInfo["AtomIndices"][ID]
            HierarchyClass = RotBondsInfo["HierarchyClass"][ID]

            MatchStatus, MatchInfo = self._MatchRotatableBondToTorsionLibrary(Mol, AtomIndices, HierarchyClass)
            self._TrackRotatableBondsAlertsInfo(RotBondsAlertsInfo, ID, AtomIndices, MatchStatus, MatchInfo)

        RotBondsAlertsStatus = self._SetupRotatableBondsAlertStatusTotalStrainEnergiesInfo(RotBondsAlertsInfo)

        return (RotBondsAlertsStatus, RotBondsAlertsInfo)


    def _InitializeRotatableBondsAlertsInfo(self):
        """Initialize alerts information for rotatable bonds."""

        RotBondsAlertsInfo = {}
        RotBondsAlertsInfo["IDs"] = []

        for DataLabel in ["RotBondsAlertsStatus", "TotalEnergy", "TotalEnergyLowerBound", "TotalEnergyUpperBound", "AnglesNotObservedCount", "MaxSingleEnergy", "MaxSingleEnergyAlertsCount"]:
            RotBondsAlertsInfo[DataLabel] = None

        for DataLabel in ["MatchStatus", "MaxSingleEnergyAlertStatus", "AtomIndices", "TorsionAtomIndices", "TorsionAngle", "HierarchyClassName", "HierarchySubClassName", "TorsionRuleNodeID", "TorsionRuleSMARTS", "EnergyMethod", "AngleNotObserved", "Energy", "EnergyLowerBound", "EnergyUpperBound"]:
            RotBondsAlertsInfo[DataLabel] = {}

        return RotBondsAlertsInfo


    def _TrackRotatableBondsAlertsInfo(self, RotBondsAlertsInfo, ID, AtomIndices, MatchStatus, MatchInfo):
        """Track alerts information for rotatable bonds."""

        if MatchInfo is None or len(MatchInfo) == 0:
            TorsionAtomIndices, TorsionAngle, HierarchyClassName, HierarchySubClassName, TorsionRuleNodeID, TorsionRuleSMARTS, EnergyMethod, AngleNotObserved, Energy, EnergyLowerBound, EnergyUpperBound = [None] * 11
        else:
            TorsionAtomIndices, TorsionAngle, HierarchyClassName, HierarchySubClassName, TorsionRuleNodeID, TorsionRuleSMARTS, EnergyMethod, AngleNotObserved, Energy, EnergyLowerBound, EnergyUpperBound = MatchInfo

        # Track torsion match information...
        RotBondsAlertsInfo["IDs"].append(ID)
        RotBondsAlertsInfo["MatchStatus"][ID] = MatchStatus
        RotBondsAlertsInfo["AtomIndices"][ID] = AtomIndices
        RotBondsAlertsInfo["TorsionAtomIndices"][ID] = TorsionAtomIndices
        RotBondsAlertsInfo["TorsionAngle"][ID] = TorsionAngle
        RotBondsAlertsInfo["HierarchyClassName"][ID] = HierarchyClassName
        RotBondsAlertsInfo["HierarchySubClassName"][ID] = HierarchySubClassName
        RotBondsAlertsInfo["TorsionRuleNodeID"][ID] = TorsionRuleNodeID
        RotBondsAlertsInfo["TorsionRuleSMARTS"][ID] = TorsionRuleSMARTS
        RotBondsAlertsInfo["EnergyMethod"][ID] = EnergyMethod
        RotBondsAlertsInfo["AngleNotObserved"][ID] = AngleNotObserved
        RotBondsAlertsInfo["Energy"][ID] = Energy
        RotBondsAlertsInfo["EnergyLowerBound"][ID] = EnergyLowerBound
        RotBondsAlertsInfo["EnergyUpperBound"][ID] = EnergyUpperBound


    def _SetupRotatableBondsAlertStatusTotalStrainEnergiesInfo(self, RotBondsAlertsInfo):
        """Setup rotatable bonds alert status along with total strain energies."""

        # Initialize...
        RotBondsAlertsStatus = False
        TotalEnergy, TotalEnergyLowerBound, TotalEnergyUpperBound, AnglesNotObservedCount  = [None, None, None, None]
        MaxSingleEnergy, MaxSingleEnergyAlertsCount  = [None, None]

        # Initialize max single energy alert status...
        for ID in RotBondsAlertsInfo["IDs"]:
            RotBondsAlertsInfo["MaxSingleEnergyAlertStatus"][ID] = None

        # Check for torsion angles not obervered in the strain library...
        AnglesNotObservedCount = 0
        for ID in RotBondsAlertsInfo["IDs"]:
            AngleNotObserved = RotBondsAlertsInfo["AngleNotObserved"][ID]
            if AngleNotObserved is not None and AngleNotObserved:
                AnglesNotObservedCount += 1

        # Setup alert status for rotable bonds...
        if AnglesNotObservedCount > 0:
            if OptionsInfo["FilterTorsionsNotObserved"]:
                RotBondsAlertsStatus = True
        else:
            TotalEnergy = 0.0
            for ID in RotBondsAlertsInfo["IDs"]:
                Energy = RotBondsAlertsInfo["Energy"][ID]
                TotalEnergy += Energy
                if self.TotalEnergyMode:
                    if TotalEnergy > self.TotalEnergyCutoff:
                        RotBondsAlertsStatus = True
                        break
                elif self.MaxSingleEnergyMode:
                    if Energy > self.MaxSingleEnergyCutoff:
                        RotBondsAlertsStatus = True
                        break
                elif self.TotalOrMaxSingleEnergyMode:
                    if TotalEnergy > self.TotalEnergyCutoff or Energy > self.MaxSingleEnergyCutoff:
                        RotBondsAlertsStatus = True
                        break

            # Setup energy infomation...
            TotalEnergy, TotalEnergyLowerBound, TotalEnergyUpperBound = [0.0, 0.0, 0.0]
            if self.MaxSingleEnergyMode or self.TotalOrMaxSingleEnergyMode:
                MaxSingleEnergy, MaxSingleEnergyAlertsCount = [0.0, 0]

            for ID in RotBondsAlertsInfo["IDs"]:
                Energy = RotBondsAlertsInfo["Energy"][ID]

                # Setup total energy along with the lower and upper bounds...
                TotalEnergy += Energy
                TotalEnergyLowerBound += RotBondsAlertsInfo["EnergyLowerBound"][ID]
                TotalEnergyUpperBound += RotBondsAlertsInfo["EnergyUpperBound"][ID]

                # Setup max single energy and max single energy alerts count...
                if self.MaxSingleEnergyMode or self.TotalOrMaxSingleEnergyMode:
                    MaxSingleEnergyAlertStatus = False

                    if Energy > MaxSingleEnergy:
                        MaxSingleEnergy = Energy
                        if Energy > self.MaxSingleEnergyCutoff:
                            MaxSingleEnergyAlertStatus = True
                            MaxSingleEnergyAlertsCount += 1

                    RotBondsAlertsInfo["MaxSingleEnergyAlertStatus"][ID] = MaxSingleEnergyAlertStatus

        RotBondsAlertsInfo["RotBondsAlertsStatus"] = RotBondsAlertsStatus

        RotBondsAlertsInfo["TotalEnergy"] = TotalEnergy
        RotBondsAlertsInfo["TotalEnergyLowerBound"] = TotalEnergyLowerBound
        RotBondsAlertsInfo["TotalEnergyUpperBound"] = TotalEnergyUpperBound

        RotBondsAlertsInfo["AnglesNotObservedCount"] = AnglesNotObservedCount

        RotBondsAlertsInfo["MaxSingleEnergy"] = MaxSingleEnergy
        RotBondsAlertsInfo["MaxSingleEnergyAlertsCount"] = MaxSingleEnergyAlertsCount

        return RotBondsAlertsStatus


    def _MatchRotatableBondToTorsionLibrary(self, Mol, RotBondAtomIndices, RotBondHierarchyClass):
        """Match rotatable bond to torsion library."""

        if TorsionAlertsUtil.IsSpecificHierarchyClass(self.TorsionLibraryInfo, RotBondHierarchyClass):
            MatchStatus, MatchInfo = self._MatchRotatableBondAgainstSpecificHierarchyClass(Mol, RotBondAtomIndices, RotBondHierarchyClass)
            if not MatchStatus:
                MatchStatus, MatchInfo = self._MatchRotatableBondAgainstGenericHierarchyClass(Mol, RotBondAtomIndices, RotBondHierarchyClass)
        else:
            MatchStatus, MatchInfo = self._MatchRotatableBondAgainstGenericHierarchyClass(Mol, RotBondAtomIndices, RotBondHierarchyClass)

        return (MatchStatus, MatchInfo)


    def _MatchRotatableBondAgainstSpecificHierarchyClass(self, Mol, RotBondAtomIndices, RotBondHierarchyClass):
        """Match rotatable bond against a specific hierarchy class."""

        TorsionLibraryInfo = self.TorsionLibraryInfo

        HierarchyClassElementNode = None
        if RotBondHierarchyClass in TorsionLibraryInfo["SpecificClasses"]["ElementNode"]:
            HierarchyClassElementNode = TorsionLibraryInfo["SpecificClasses"]["ElementNode"][RotBondHierarchyClass]

        if HierarchyClassElementNode is None:
            return (False, None, None, None)

        TorsionAlertsUtil.TrackHierarchyClassElementNode(TorsionLibraryInfo, HierarchyClassElementNode)
        MatchStatus, MatchInfo = self._ProcessElementForRotatableBondMatch(Mol, RotBondAtomIndices, HierarchyClassElementNode)
        TorsionAlertsUtil.RemoveLastHierarchyClassElementNodeFromTracking(TorsionLibraryInfo)

        return (MatchStatus, MatchInfo)


    def _MatchRotatableBondAgainstGenericHierarchyClass(self, Mol, RotBondAtomIndices, RotBondHierarchyClass):
        """Match rotatable bond against a generic hierarchy class."""

        TorsionLibraryInfo = self.TorsionLibraryInfo

        HierarchyClassElementNode = TorsionAlertsUtil.GetGenericHierarchyClassElementNode(TorsionLibraryInfo)
        if HierarchyClassElementNode is None:
            return (False, None)

        TorsionAlertsUtil.TrackHierarchyClassElementNode(TorsionLibraryInfo, HierarchyClassElementNode)

        #  Match hierarchy subclasses before matching torsion rules...
        MatchStatus, MatchInfo = self._MatchRotatableBondAgainstGenericHierarchySubClasses(Mol, RotBondAtomIndices, HierarchyClassElementNode)

        if not MatchStatus:
            MatchStatus, MatchInfo = self._MatchRotatableBondAgainstGenericHierarchyTorsionRules(Mol, RotBondAtomIndices, HierarchyClassElementNode)

        TorsionAlertsUtil.RemoveLastHierarchyClassElementNodeFromTracking(TorsionLibraryInfo)

        return (MatchStatus, MatchInfo)


    def _MatchRotatableBondAgainstGenericHierarchySubClasses(self, Mol, RotBondAtomIndices, HierarchyClassElementNode):
        """Match rotatable bond againat generic hierarchy subclasses."""

        for ElementChildNode in HierarchyClassElementNode:
            if ElementChildNode.tag != "hierarchySubClass":
                continue

            SubClassMatchStatus = self._ProcessHierarchySubClassElementForRotatableBondMatch(Mol, RotBondAtomIndices, ElementChildNode)

            if SubClassMatchStatus:
                MatchStatus, MatchInfo = self._ProcessElementForRotatableBondMatch(Mol, RotBondAtomIndices, ElementChildNode)

                if MatchStatus:
                    return (MatchStatus, MatchInfo)

        return(False, None)


    def _MatchRotatableBondAgainstGenericHierarchyTorsionRules(self, Mol, RotBondAtomIndices, HierarchyClassElementNode):
        """Match rotatable bond againat torsion rules generic hierarchy class."""

        for ElementChildNode in HierarchyClassElementNode:
            if ElementChildNode.tag != "torsionRule":
                continue

            MatchStatus, MatchInfo = self._ProcessTorsionRuleElementForRotatableBondMatch(Mol, RotBondAtomIndices, ElementChildNode)

            if MatchStatus:
                return (MatchStatus, MatchInfo)

        return(False, None)


    def _ProcessElementForRotatableBondMatch(self, Mol, RotBondAtomIndices, ElementNode):
        """Process element node to recursively match rotatable bond against hierarchy
        subclasses and torsion rules."""

        TorsionLibraryInfo = self.TorsionLibraryInfo

        for ElementChildNode in ElementNode:
            if ElementChildNode.tag == "hierarchySubClass":
                SubClassMatchStatus = self._ProcessHierarchySubClassElementForRotatableBondMatch(Mol, RotBondAtomIndices, ElementChildNode)

                if SubClassMatchStatus:
                    TorsionAlertsUtil.TrackHierarchySubClassElementNode(TorsionLibraryInfo, ElementChildNode)

                    MatchStatus, MatchInfo = self._ProcessElementForRotatableBondMatch(Mol, RotBondAtomIndices, ElementChildNode)
                    if MatchStatus:
                        TorsionAlertsUtil.RemoveLastHierarchySubClassElementNodeFromTracking(TorsionLibraryInfo)
                        return (MatchStatus, MatchInfo)

                    TorsionAlertsUtil.RemoveLastHierarchySubClassElementNodeFromTracking(TorsionLibraryInfo)

            elif ElementChildNode.tag == "torsionRule":
                MatchStatus, MatchInfo = self._ProcessTorsionRuleElementForRotatableBondMatch(Mol, RotBondAtomIndices, ElementChildNode)

                if MatchStatus:
                    return (MatchStatus, MatchInfo)

        return (False, None)


    def _ProcessHierarchySubClassElementForRotatableBondMatch(self, Mol, RotBondAtomIndices, ElementNode):
        """Process hierarchy subclass element to match rotatable bond."""

        # Setup subclass SMARTS pattern mol...
        SubClassPatternMol = TorsionAlertsUtil.SetupHierarchySubClassElementPatternMol(self.TorsionLibraryInfo, ElementNode)
        if SubClassPatternMol is None:
            return False

        # Match SMARTS pattern...
        SubClassPatternMatches = TorsionAlertsUtil.FilterSubstructureMatchesByAtomMapNumbers(Mol, SubClassPatternMol, Mol.GetSubstructMatches(SubClassPatternMol, useChirality = False))
        if len(SubClassPatternMatches) == 0:
            return False

        # Match rotatable bond indices...
        RotBondAtomIndex1, RotBondAtomIndex2 = RotBondAtomIndices
        MatchStatus = False
        for SubClassPatternMatch in SubClassPatternMatches:
            if len(SubClassPatternMatch) == 2:
                # Matched to pattern containing map atom numbers ":2" and ":3"...
                CentralAtomsIndex1, CentralAtomsIndex2 = SubClassPatternMatch
            elif len(SubClassPatternMatch) == 4:
                # Matched to pattern containing map atom numbers ":1", ":2", ":3" and ":4"...
                CentralAtomsIndex1 = SubClassPatternMatch[1]
                CentralAtomsIndex2 = SubClassPatternMatch[2]
            elif len(SubClassPatternMatch) == 3:
                SubClassSMARTSPattern = ElementNode.get("smarts")
                if TorsionAlertsUtil.DoesSMARTSContainsMappedAtoms(SubClassSMARTSPattern, [":2", ":3", ":4"]):
                    # Matched to pattern containing map atom numbers ":2", ":3" and ":4"...
                    CentralAtomsIndex1 = SubClassPatternMatch[0]
                    CentralAtomsIndex2 = SubClassPatternMatch[1]
                else:
                    # Matched to pattern containing map atom numbers ":1", ":2" and ":3"...
                    CentralAtomsIndex1 = SubClassPatternMatch[1]
                    CentralAtomsIndex2 = SubClassPatternMatch[2]
            else:
                continue

            if CentralAtomsIndex1 != CentralAtomsIndex2:
                if ((CentralAtomsIndex1 == RotBondAtomIndex1 and CentralAtomsIndex2 == RotBondAtomIndex2) or (CentralAtomsIndex1 == RotBondAtomIndex2 and CentralAtomsIndex2 == RotBondAtomIndex1)):
                    MatchStatus = True
                    break

        return (MatchStatus)


    def _ProcessTorsionRuleElementForRotatableBondMatch(self, Mol, RotBondAtomIndices, ElementNode):
        """Process torsion rule element to match rotatable bond."""

        TorsionLibraryInfo = self.TorsionLibraryInfo

        #  Retrieve torsions matched to rotatable bond...
        TorsionAtomIndicesList, TorsionAnglesList = self._MatchTorsionRuleToRotatableBond(Mol, RotBondAtomIndices, ElementNode)
        if TorsionAtomIndicesList is None:
            return (False, None)

        # Setup torsion angles and enery bin information for matched torsion rule...
        TorsionRuleAnglesInfo = TorsionAlertsUtil.SetupTorsionRuleAnglesInfo(TorsionLibraryInfo, ElementNode)
        if TorsionRuleAnglesInfo is None:
            return (False, None)

        # Setup highest strain energy for matched torsions...
        TorsionAtomIndices, TorsionAngle, EnergyMethod, AngleNotObserved, Energy, EnergyLowerBound, EnergyUpperBound = self._SelectHighestStrainEnergyTorsionForRotatableBond(TorsionRuleAnglesInfo, TorsionAtomIndicesList, TorsionAnglesList)

        # Setup hierarchy class and subclass names...
        HierarchyClassName, HierarchySubClassName = TorsionAlertsUtil.SetupHierarchyClassAndSubClassNamesForRotatableBond(TorsionLibraryInfo)

        # Setup rule node ID...
        TorsionRuleNodeID = ElementNode.get("NodeID")

        # Setup SMARTS...
        TorsionRuleSMARTS = ElementNode.get("smarts")
        if " " in TorsionRuleSMARTS:
            TorsionRuleSMARTS = TorsionRuleSMARTS.replace(" ", "")

        # Setup match info...
        MatchInfo = [TorsionAtomIndices, TorsionAngle, HierarchyClassName, HierarchySubClassName, TorsionRuleNodeID, TorsionRuleSMARTS, EnergyMethod, AngleNotObserved, Energy, EnergyLowerBound, EnergyUpperBound]

        # Setup match status...
        MatchStatus = True

        return (MatchStatus, MatchInfo)


    def _SelectHighestStrainEnergyTorsionForRotatableBond(self, TorsionRuleAnglesInfo, TorsionAtomIndicesList, TorsionAnglesList):
        """Select highest strain energy torsion matched to a rotatable bond."""

        TorsionAtomIndices, TorsionAngle, EnergyMethod, AngleNotObserved, Energy, EnergyLowerBound, EnergyUpperBound = [None] * 7
        ValidEnergyValue, ValidCurrentEnergyValue = [False] * 2

        FirstTorsion = True
        for Index in range(0, len(TorsionAtomIndicesList)):
            CurrentTorsionAtomIndices = TorsionAtomIndicesList[Index]
            CurrentTorsionAngle = TorsionAnglesList[Index]

            if FirstTorsion:
                FirstTorsion = False
                EnergyMethod, AngleNotObserved, Energy, EnergyLowerBound, EnergyUpperBound = self._SetupStrainEnergyForRotatableBond(TorsionRuleAnglesInfo, CurrentTorsionAngle)
                TorsionAtomIndices = CurrentTorsionAtomIndices
                TorsionAngle = CurrentTorsionAngle
                ValidEnergyValue = self._IsEnergyValueValid(Energy)
                continue

            # Select highest strain energy...
            CurrentEnergyMethod, CurrentAngleNotObserved, CurrentEnergy, CurrentEnergyLowerBound, CurrentEnergyUpperBound = self._SetupStrainEnergyForRotatableBond(TorsionRuleAnglesInfo, CurrentTorsionAngle)
            ValidCurrentEnergyValue = self._IsEnergyValueValid(CurrentEnergy)

            UpdateValues = False
            if ValidEnergyValue and ValidCurrentEnergyValue:
                if CurrentEnergy > Energy:
                    UpdateValues = True
            elif ValidCurrentEnergyValue:
                if not ValidEnergyValue:
                    UpdateValues = True

            if UpdateValues:
                EnergyMethod, AngleNotObserved, Energy, EnergyLowerBound, EnergyUpperBound = [CurrentEnergyMethod, CurrentAngleNotObserved, CurrentEnergy, CurrentEnergyLowerBound, CurrentEnergyUpperBound]
                TorsionAtomIndices = CurrentTorsionAtomIndices
                TorsionAngle = CurrentTorsionAngle

        return (TorsionAtomIndices, TorsionAngle, EnergyMethod, AngleNotObserved, Energy, EnergyLowerBound, EnergyUpperBound)


    def _IsEnergyValueValid(self, Value):
        """Check for valid energy value."""

        return False if (Value is None or math.isnan(Value) or math.isinf(Value)) else True


    def _SetupStrainEnergyForRotatableBond(self, TorsionRuleAnglesInfo, TorsionAngle):
        """Setup strain energy for rotatable bond."""

        if TorsionRuleAnglesInfo["EnergyMethodExact"]:
            return (self._SetupStrainEnergyForRotatableBondByExactMethod(TorsionRuleAnglesInfo, TorsionAngle))
        elif TorsionRuleAnglesInfo["EnergyMethodApproximate"]:
            return (self._SetupStrainEnergyForRotatableBondByApproximateMethod(TorsionRuleAnglesInfo, TorsionAngle))
        else:
            EnergyMethod, AngleNotObserved, Energy, EnergyLowerBound, EnergyUpperBound = [None, None, None, None, None]
            return (EnergyMethod, AngleNotObserved, Energy, EnergyLowerBound, EnergyUpperBound)


    def _SetupStrainEnergyForRotatableBondByExactMethod(self, TorsionRuleAnglesInfo, TorsionAngle):
        """Setup strain energy for rotatable bond by exact method."""

        EnergyMethod, AngleNotObserved, Energy, EnergyLowerBound, EnergyUpperBound = ["Exact", None, None, None, None]

        # Map angle to energy bin numbers...
        BinNum = math.ceil(TorsionAngle / 10) + 17
        PreviousBinNum = (BinNum + 35) % 36

        # Bin angle from -170 to 180 by the right end points...
        BinAngleRightSide = (BinNum - 17) * 10

        # Angle offset towards the left of the bin from the right end point...
        AngleOffset = TorsionAngle - BinAngleRightSide

        BinEnergy = TorsionRuleAnglesInfo["HistogramEnergy"][BinNum]
        PreviousBinEnergy = TorsionRuleAnglesInfo["HistogramEnergy"][PreviousBinNum]
        Energy =  BinEnergy + (BinEnergy - PreviousBinEnergy)/10.0 * AngleOffset

        BinEnergyLowerBound = TorsionRuleAnglesInfo["HistogramEnergyLowerBound"][BinNum]
        PreviousBinEnergyLowerBound = TorsionRuleAnglesInfo["HistogramEnergyLowerBound"][PreviousBinNum]
        EnergyLowerBound =  BinEnergyLowerBound + (BinEnergyLowerBound - PreviousBinEnergyLowerBound)/10.0 * AngleOffset

        BinEnergyUpperBound = TorsionRuleAnglesInfo["HistogramEnergyUpperBound"][BinNum]
        PreviousBinEnergyUpperBound = TorsionRuleAnglesInfo["HistogramEnergyUpperBound"][PreviousBinNum]
        EnergyUpperBound =  BinEnergyUpperBound + (BinEnergyUpperBound - PreviousBinEnergyUpperBound)/10.0 * AngleOffset

        return (EnergyMethod, AngleNotObserved, Energy, EnergyLowerBound, EnergyUpperBound)


    def _SetupStrainEnergyForRotatableBondByApproximateMethod(self, TorsionRuleAnglesInfo, TorsionAngle):
        """Setup strain energy for rotatable bond by approximate method."""

        EnergyMethod, AngleNotObserved, Energy, EnergyLowerBound, EnergyUpperBound = ["Approximate", True, None, None, None]

        for AngleID in TorsionRuleAnglesInfo["IDs"]:
            Tolerance2 = TorsionRuleAnglesInfo["Tolerance2"][AngleID]
            Theta0 = TorsionRuleAnglesInfo["Theta0"][AngleID]

            AngleDiff = TorsionAlertsUtil.CalculateTorsionAngleDifference(TorsionAngle, Theta0)
            if abs(AngleDiff) <= Tolerance2:
                Beta1 = TorsionRuleAnglesInfo["Beta1"][AngleID]
                Beta2 = TorsionRuleAnglesInfo["Beta2"][AngleID]

                Energy = Beta1*(AngleDiff ** 2) + Beta2*(AngleDiff** 4)

                # Estimates of lower and upper bound are not available for
                # approximate method...
                EnergyLowerBound = Energy
                EnergyUpperBound = Energy

                AngleNotObserved = False

                break

        return (EnergyMethod, AngleNotObserved, Energy, EnergyLowerBound, EnergyUpperBound)


    def _MatchTorsionRuleToRotatableBond(self, Mol, RotBondAtomIndices, ElementNode):
        """Retrieve matched torsions for torsion rule matched to rotatable bond."""

        # Get torsion matches...
        TorsionMatches = self._GetMatchesForTorsionRule(Mol, ElementNode)
        if TorsionMatches is None or len(TorsionMatches) == 0:
            return (None, None)

        # Identify all torsion matches corresponding to central atoms in RotBondAtomIndices...
        RotBondAtomIndex1, RotBondAtomIndex2 = RotBondAtomIndices
        RotBondTorsionMatches, RotBondTorsionAngles = [None] * 2

        for TorsionMatch in TorsionMatches:
            CentralAtomIndex1 = TorsionMatch[1]
            CentralAtomIndex2 = TorsionMatch[2]

            if ((CentralAtomIndex1 == RotBondAtomIndex1 and CentralAtomIndex2 == RotBondAtomIndex2) or (CentralAtomIndex1 == RotBondAtomIndex2 and CentralAtomIndex2 == RotBondAtomIndex1)):
                TorsionAngle = self._CalculateTorsionAngle(Mol, TorsionMatch)
                if RotBondTorsionMatches is None:
                    RotBondTorsionMatches = []
                    RotBondTorsionAngles = []
                RotBondTorsionMatches.append(TorsionMatch)
                RotBondTorsionAngles.append(TorsionAngle)

        return (RotBondTorsionMatches, RotBondTorsionAngles)


    def _CalculateTorsionAngle(self, Mol, TorsionMatch):
        """Calculate torsion angle."""

        # Calculate torsion angle using torsion atom indices..
        MolConf = Mol.GetConformer(0)
        TorsionAngle = rdMolTransforms.GetDihedralDeg(MolConf, TorsionMatch[0], TorsionMatch[1], TorsionMatch[2], TorsionMatch[3])
        TorsionAngle = round(TorsionAngle, 2)

        return TorsionAngle


    def _GetMatchesForTorsionRule(self, Mol, ElementNode):
        """Get matches for torsion rule."""

        # Match torsions...
        TorsionMatches = self._GetSubstructureMatchesForTorsionRule(Mol, ElementNode)

        if TorsionMatches is None or len(TorsionMatches) == 0:
            return TorsionMatches

        # Filter torsion matches...
        FiltertedTorsionMatches = []
        for TorsionMatch in TorsionMatches:
            if len(TorsionMatch) != 4:
                continue

            # Ignore matches containing hydrogen atoms as first or last atom...
            if Mol.GetAtomWithIdx(TorsionMatch[0]).GetAtomicNum() == 1:
                continue
            if Mol.GetAtomWithIdx(TorsionMatch[3]).GetAtomicNum() == 1:
                continue

            FiltertedTorsionMatches.append(TorsionMatch)

        return FiltertedTorsionMatches


    def _GetSubstructureMatchesForTorsionRule(self, Mol, ElementNode):
        """Get substructure matches for a torsion rule."""

        # Setup torsion rule SMARTS pattern mol....
        TorsionRuleNodeID = ElementNode.get("NodeID")
        TorsionSMARTSPattern = ElementNode.get("smarts")
        TorsionPatternMol = TorsionAlertsUtil.SetupTorsionRuleElementPatternMol(self.TorsionLibraryInfo, ElementNode, TorsionRuleNodeID, TorsionSMARTSPattern)
        if TorsionPatternMol is None:
            return None

        # Match torsions...
        TorsionMatches = TorsionAlertsUtil.FilterSubstructureMatchesByAtomMapNumbers(Mol, TorsionPatternMol, Mol.GetSubstructMatches(TorsionPatternMol, useChirality = False))

        return TorsionMatches
